"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import jax
from jax import make_jaxpr
from jax.extend.core import Jaxpr, Literal, ClosedJaxpr
from jax.tree_util import tree_flatten
from jax.errors import UnexpectedTracerError

from qrisp.jasp.jasp_expression import invert_jaspr, collect_environments
from qrisp.jasp import (
    eval_jaxpr,
    pjit_to_gate,
    flatten_environments,
    cond_to_cl_control,
    extract_invalues,
    insert_outvalues
)
from qrisp.jasp.primitives import AbstractQuantumCircuit, QuantumPrimitive


class Jaspr(ClosedJaxpr):
    """
    The ``Jaspr`` class enables an efficient representations of a wide variety
    of (hybrid) algorithms. For many applications, the representation is agnostic
    to the scale of the problem, implying function calls with 10 or 10000 qubits
    can be represented by the same object. The actual unfolding to a circuit-level
    description is outsourced to
    `established, classical compilation infrastructure <https://mlir.llvm.org/>`_,
    implying state-of-the-art compilation speed can be reached.

    As a subtype of ``jax.extend.core.ClosedJaxpr``, Jasprs are embedded into the well matured
    `Jax ecosystem <https://github.com/n2cholas/awesome-jax>`_,
    which facilitates the compilation of classical `real-time computation <https://arxiv.org/abs/2206.12950>`_
    using some of the most advanced libraries in the world such as
    `CUDA <https://jax.readthedocs.io/en/latest/Custom_Operation_for_GPUs.html>`_.
    Especially `machine learning <https://ai.google.dev/gemma/docs/jax_inference>`_
    and other scientific computations tasks are particularly well supported.

    To get a better understanding of the syntax and semantics of Jaxpr (and with
    that also Jaspr) please check `this link <https://jax.readthedocs.io/en/latest/jaxpr.html>`__.

    Similar to Jaxpr, Jaspr objects represent (hybrid) quantum
    algorithms in the form of a `functional programming language <https://en.wikipedia.org/wiki/Functional_programming>`_
    in `SSA-form <https://en.wikipedia.org/wiki/Static_single-assignment_form>`_.

    It is possible to compile Jaspr objects into QIR, which is facilitated by the
    `Catalyst framework <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__
    (check :meth:`qrisp.jasp.jaspr.to_qir` for more details).

    Qrisp scripts can be turned into Jaspr objects by
    calling the ``make_jaspr`` function, which has similar semantics as
    `jax.make_jaxpr <https://jax.readthedocs.io/en/latest/_autosummary/jax.make_jaxpr.html>`_.

    ::

        from qrisp import *
        from qrisp.jasp import make_jaspr

        def test_fun(i):

            qv = QuantumFloat(i, -1)
            x(qv[0])
            cx(qv[0], qv[i-1])
            meas_res = measure(qv)
            meas_res += 1
            return meas_res


        jaspr = make_jaspr(test_fun)(4)
        print(jaspr)

    This will give you the following output:

    .. code-block::

        { lambda ; a:QuantumCircuit b:i32[]. let
            c:QuantumCircuit d:QubitArray = create_qubits a b
            e:Qubit = get_qubit d 0
            f:QuantumCircuit = x c e
            g:i32[] = sub b 1
            h:Qubit = get_qubit d g
            i:QuantumCircuit = cx f e h
            j:QuantumCircuit k:i32[] = measure i d
            l:f32[] = convert_element_type[new_dtype=float64 weak_type=True] k
            m:f32[] = mul l 0.5
            n:f32[] = add m 1.0
          in (j, n) }


    A defining feature of the Jaspr class is that the first input and the
    first output are always of QuantumCircuit type. Therefore, Jaspr objects always
    represent some (hybrid) quantum operation.

    Qrisp comes with a built-in Jaspr interpreter. For that you simply have to
    call the object like a function:


    >>> print(jaspr(2))
    2.5
    >>> print(jaspr(4))
    5.5
    """

    __slots__ = (
        "permeability",
        "isqfree",
        "hashvalue",
        "ctrl_jaspr",
        "inv_jaspr",
        "envs_flattened",
    )

    def __init__(
        self, *args, permeability=None, isqfree=None, ctrl_jaspr=None, inv_jaspr=None, **kwargs
    ):

        if len(args) == 2:
            if not isinstance(args[0], Jaxpr) or not isinstance(args[1], list):
                raise Exception(f"Tried to call the Jaspr constructor with two arguments and signature {type(args[0]), type(args[1])} (allowed is (Jaxpr, list))")
            kwargs["jaxpr"] = args[0]            
            kwargs["consts"] = args[1]
            
        elif len(args) == 1:
            if not isinstance(args[0], ClosedJaxpr):
                raise Exception(f"Tried to call the Jaspr constructor with one argument and signature {type(args[0])} (allowed is (ClosedJaxpr))")
            kwargs["jaxpr"] = args[0].jaxpr
            kwargs["consts"] = args[0].consts

        if "jaxpr" in kwargs:

            ClosedJaxpr.__init__(self,
                                 kwargs["jaxpr"],
                                 kwargs["consts"])
        else:
            
            if "consts" in kwargs:
                consts = kwargs["consts"]
                del kwargs["consts"]
            else:
                if len(kwargs["constvars"]):
                    raise Exception("Tried to create Jaspr with constvars but no constants")
                consts = []
            
            ClosedJaxpr.__init__(self, 
                                 jaxpr = Jaxpr(**kwargs), 
                                 consts = consts)

        self.hashvalue = id(self)
        self.permeability = {}
        if permeability is None:
            permeability = {}
        for var in self.constvars + self.invars + self.outvars:
            if isinstance(var, Literal):
                continue
            self.permeability[var] = permeability.get(var, None)

        self.isqfree = isqfree
        self.ctrl_jaspr = ctrl_jaspr
        self.inv_jaspr = inv_jaspr
        self.envs_flattened = False

        if not isinstance(self.invars[-1].aval, AbstractQuantumCircuit):
            raise Exception(
                f"Tried to create a Jaspr from data that doesn't have a QuantumCircuit the last argument (got {type(self.invars[-1].aval)} instead)"
            )

        if not isinstance(self.outvars[-1].aval, AbstractQuantumCircuit):
            raise Exception(
                f"Tried to create a Jaspr from data that doesn't have a QuantumCircuit the last entry of return type (got {type(self.outvars[-1].aval)} instead)"
            )

    @property
    def constvars(self):
        return self.jaxpr.constvars
    
    @property
    def eqns(self):
        return self.jaxpr.eqns

    @property
    def invars(self):
        return self.jaxpr.invars

    @property
    def outvars(self):
        return self.jaxpr.outvars
    
    @property
    def debug_info(self):
        return self.jaxpr.debug_info

    def __hash__(self):
        return self.hashvalue

    def __eq__(self, other):
        if not isinstance(other, Jaxpr):
            return False

        return id(self) == id(other)

    def copy(self):

        if self.ctrl_jaspr is None:
            ctrl_jaspr = None
        else:
            ctrl_jaspr = self.ctrl_jaspr.copy()

        res = Jaspr(
            permeability=self.permeability,
            isqfree=self.isqfree,
            ctrl_jaspr=ctrl_jaspr,
            constvars=list(self.constvars),
            invars=list(self.invars),
            outvars=list(self.outvars),
            eqns=list(self.eqns),
            effects=self.effects,
            debug_info=self.debug_info,
        )

        res.envs_flattened = self.envs_flattened

        return res

    def inverse(self):
        """
        Returns the inverse Jaspr (if applicable). For Jaspr that contain realtime
        computations or measurements, the inverse does not exist.

        Returns
        -------
        Jaspr
            The daggered Jaspr.

        Examples
        --------

        We create a simple script and inspect the daggered version:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.inverse())
            # Yields
            # { lambda ; a:QuantumCircuit b:i32[]. let
            #     c:QuantumCircuit d:QubitArray = create_qubits a b
            #     e:Qubit = get_qubit d 0
            #     f:Qubit = get_qubit d 1
            #     g:QuantumCircuit = t_dg c f
            #     h:QuantumCircuit = cx g e f
            #   in (h, d) }
        """
        return invert_jaspr(self)

    def control(self, num_ctrl, ctrl_state=-1):
        """
        Returns the controlled version of the Jaspr. The control qubits are added
        to the signature of the Jaspr as the arguments after the QuantumCircuit.

        Parameters
        ----------
        num_ctrl : int
            The amount of controls to be added.
        ctrl_state : int of str, optional
            The control state on which to activate. The default is -1.

        Returns
        -------
        Jaspr
            The controlled Jaspr.

        Examples
        --------

        We create a simple script and inspect the controlled version:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.control(2))
            # Yields
            # { lambda ; a:QuantumCircuit b:Qubit c:Qubit d:i32[]. let
            #     e:QuantumCircuit f:QubitArray = create_qubits a 1
            #     g:Qubit = get_qubit f 0
            #     h:QuantumCircuit = ccx e b c g
            #     i:QuantumCircuit j:QubitArray = create_qubits h d
            #     k:Qubit = get_qubit j 0
            #     l:Qubit = get_qubit j 1
            #     m:QuantumCircuit = ccx i g k l
            #     n:QuantumCircuit = ct m g l
            #     o:QuantumCircuit = ccx n b c g
            #   in (o, j) }

        We see that the control qubits are part of the function signature
        (``a`` and ``b``)

        """
        if self.ctrl_jaspr is not None and num_ctrl == 1 and ctrl_state == -1:
            return self.ctrl_jaspr

        from qrisp.jasp import ControlledJaspr

        if isinstance(ctrl_state, int):
            if ctrl_state < 0:
                ctrl_state += 2**num_ctrl

            ctrl_state = bin(ctrl_state)[2:].zfill(num_ctrl)
        else:
            ctrl_state = str(ctrl_state)

        return ControlledJaspr.from_cache(self, ctrl_state)

    def to_qc(self, *args):
        """
        Converts the Jaspr into a :ref:`QuantumCircuit` if applicable. Circuit
        conversion of algorithms involving realtime computations is not possible.
        
        Any computations that perform classical postprocessing of measurements
        can not be reflected within the QuantumCircuit object itself and will
        generate an object of type ``ProcessedMeasurement``. These objects hold
        no further information and are simply used as placeholders to emulate
        the computation.

        Parameters
        ----------
        *args : tuple
            The arguments to call the Jaspr with.

        Returns
        -------
        return_values : tuple
            The return values of the Jaspr. QuantumVariable return types are
            returned as lists of Qubits.
        :ref:`QuantumCircuit`
            The resulting QuantumCircuit.

        Examples
        --------

        We create a simple script and inspect the QuantumCircuit:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(example_function)(2)

            qb_list, qc = jaspr.to_qc(2)
            print(qc)
            # Yields
            # qb_0: ──■───────
            #       ┌─┴─┐┌───┐
            # qb_1: ┤ X ├┤ T ├
            #       └───┘└───┘
            
        To demonstrate the behavior under measurement post-processing, we build
        a similar script:
            
        ::
            
            from qrisp import ProcessedMeasurement
            
            def example_function(i):
            
                qf = QuantumFloat(i)
                cx(qf[0], qf[1])
                t(qf[1])
                
                meas_res = measure(qf)
                # Perform classical post processing
                meas_res *= 2
                return meas_res
            
            jaspr = make_jaspr(example_function)(2)
            
            meas_res, qc = jaspr.to_qc(2)
            print(isinstance(meas_res, ProcessedMeasurement))
            # True
            
        

        """
        from qrisp import QuantumCircuit, Clbit

        jaspr = self
        
        def eqn_evaluator(eqn, context_dic):
            if eqn.primitive.name == "pjit" and isinstance(
                eqn.params["jaxpr"], Jaspr
            ):
                return pjit_to_gate(eqn, context_dic, eqn_evaluator)
            elif eqn.primitive.name == "cond":
                return cond_to_cl_control(eqn, context_dic, eqn_evaluator)
            elif eqn.primitive.name == "convert_element_type":
                if isinstance(context_dic[eqn.invars[0]], (ProcessedMeasurement, Clbit)):
                    context_dic[eqn.outvars[0]] = context_dic[eqn.invars[0]]
                    return
                elif isinstance(context_dic[eqn.invars[0]], list) and isinstance(
                    context_dic[eqn.invars[0]][0], (ProcessedMeasurement, Clbit)
                ):
                    context_dic[eqn.outvars[0]] = context_dic[eqn.invars[0]]
                    return
                return True
            else:

                invalues = extract_invalues(eqn, context_dic)
                for val in invalues:
                    if isinstance(val, list) and len(val):
                        if isinstance(val[0], (ProcessedMeasurement, Clbit)):
                            break
                    elif isinstance(val, (ProcessedMeasurement, Clbit)):
                        break
                else:
                    if isinstance(eqn.primitive, QuantumPrimitive):
                        outvalues = eqn.primitive.impl(*invalues, **eqn.params)
                        insert_outvalues(eqn, context_dic, outvalues)
                        return
                    else:
                        return True
                
            if len(eqn.outvars) == 0:
                return
            elif len(eqn.outvars) == 1 and not eqn.primitive.multiple_results:
                outvalues = ProcessedMeasurement()
            elif len(eqn.outvars) >= 1:
                outvalues = [ProcessedMeasurement() for _ in range(len(eqn.outvars))]
            
            insert_outvalues(eqn, context_dic, outvalues)

        ammended_args = list(args) + [QuantumCircuit()]
        
        if len(ammended_args) != len(jaspr.invars):
            raise Exception(
                "Supplied invalid number of arguments to Jaspr.to_qc (please exclude any static arguments, in particular callables)"
            )

        res = eval_jaxpr(jaspr, eqn_evaluator=eqn_evaluator)(*(ammended_args))

        return res

    def eval(self, *args, eqn_evaluator=lambda x, y: True):
        return eval_jaxpr(self, eqn_evaluator=eqn_evaluator)(*args)

    def flatten_environments(self):
        """
        Flattens all environments by applying the corresponding compilation
        routines such that no more ``q_env`` primitives are left.

        Returns
        -------
        Jaspr
            The Jaspr with flattened environments.

        Examples
        --------

        We create a Jaspr containing an :ref:`InversionEnvironment` and flatten:

        ::

                def test_function(i):
                        qv = QuantumVariable(i)

                        with invert():
                                t(qv[0])
                                cx(qv[0], qv[1])

                        return qv

                jaspr = make_jaspr(test_function)(2)
                print(jaspr)

        ::

                { lambda ; a:QuantumCircuit b:i32[]. let
                        c:QuantumCircuit d:QubitArray = create_qubits a b
                        e:QuantumCircuit = q_env[
                        jaspr={ lambda ; f:QuantumCircuit d:QubitArray. let
                  g:Qubit = get_qubit d 0
                  h:QuantumCircuit = t f g
                  i:Qubit = get_qubit d 1
                  j:QuantumCircuit = cx h g i
                in (j,) }
                        type=InversionEnvironment
                        ] c d
                  in (e, d) }

        You can see how the body of the :ref:`InversionEnvironment` is __collected__
        into another Jaspr. This reflects the fact that at their core,
        :ref:`QuantumEnvironment <QuantumEnvironment>` describe `higher-order
        quantum functions <https://en.wikipedia.org/wiki/Higher-order_function>`_
        (ie. functions that operate on functions). In order to apply the
        transformations induced by the QuantumEnvironment, we can call
        ``jaspr.flatten_environments``:

        >>> print(jaspr.flatten_environments)
        { lambda ; a:QuantumCircuit b:i32[]. let
            c:QuantumCircuit d:QubitArray = create_qubits a b
            e:Qubit = get_qubit d 0
            f:Qubit = get_qubit d 1
            g:QuantumCircuit = cx c e f
            h:QuantumCircuit = t_dg g e
          in (h, d) }

        We see that as expected, the order of the ``cx`` and the ``t`` gate has been switched and the ``t`` gate has been turned into a ``t_dg``.

        """
        res = flatten_environments(self)
        if self.ctrl_jaspr is not None:
            res.ctrl_jaspr = self.ctrl_jaspr.flatten_environments()
        if self.inv_jaspr is not None:
            if not self.inv_jaspr.envs_flattened:
                res.inv_jaspr = self.inv_jaspr.flatten_environments()
            else:
                res.inv_jaspr = self.inv_jaspr
            res.inv_jaspr.inv_jaspr = res
        return res

    def __call__(self, *args):
        from qrisp.jasp.evaluation_tools.jaspification import simulate_jaspr

        return simulate_jaspr(self, *args)

        if len(self.outvars) == 1:
            return None

        from qrisp.simulator import BufferedQuantumState

        args = [BufferedQuantumState()] + list(tree_flatten(args)[0])

        from qrisp.jasp import extract_invalues, insert_outvalues, eval_jaxpr

        flattened_jaspr = self

        def eqn_evaluator(eqn, context_dic):
            if eqn.primitive.name == "pjit":

                if eqn.params["name"] == "expectation_value_eval_function":
                    from qrisp.jasp.program_control import sampling_evaluator

                    sampling_evaluator("ev")(
                        eqn, context_dic, eqn_evaluator=eqn_evaluator
                    )
                    return

                if eqn.params["name"] == "sampling_eval_function":
                    from qrisp.jasp.program_control import sampling_evaluator

                    sampling_evaluator("array")(
                        eqn, context_dic, eqn_evaluator=eqn_evaluator
                    )
                    return

                invalues = extract_invalues(eqn, context_dic)
                outvalues = eval_jaxpr(
                    eqn.params["jaxpr"], eqn_evaluator=eqn_evaluator
                )(*invalues)
                if not isinstance(outvalues, (list, tuple)):
                    outvalues = [outvalues]
                insert_outvalues(eqn, context_dic, outvalues)
            elif eqn.primitive.name == "jasp.create_quantum_kernel":
                insert_outvalues(eqn, context_dic, BufferedQuantumState())
            else:
                return True

        res = eval_jaxpr(flattened_jaspr, eqn_evaluator=eqn_evaluator)(
            *(args + self.consts)
        )

        if len(self.outvars) == 2:
            return res[1]
        else:
            return res[1:]

    def inline(self, *args):

        from qrisp.jasp import TracingQuantumSession

        qs = TracingQuantumSession.get_instance()
        abs_qc = qs.abs_qc

        ammended_args = list(args) + [abs_qc]
        res = eval_jaxpr(self)(*ammended_args)

        if isinstance(res, tuple):
            new_abs_qc = res[-1]
            res = res[:-1]
        else:
            new_abs_qc = res
            res = None
        qs.abs_qc = new_abs_qc
        return res

    def count_ops(self, *args, meas_behavior="1"):
        from qrisp.jasp.evaluation_tools import profile_jaspr

        return profile_jaspr(self, meas_behavior)(*args)

    def embedd(self, *args, name=None, inline=False):
        from qrisp.jasp import TracingQuantumSession, get_last_equation

        qs = TracingQuantumSession.get_instance()
        abs_qc = qs.abs_qc

        ammended_args = list(args) + [abs_qc]
        if not inline:
            res = jax.jit(eval_jaxpr(self))(*ammended_args)

            eqn = get_last_equation()
            
            eqn.params["jaxpr"] = self
            if name is not None:
                eqn.params["name"] = name
        else:
            res = eval_jaxpr(self)(*ammended_args)

        if isinstance(res, tuple):
            new_abs_qc = res[-1]
            res = res[:-1]
        else:
            new_abs_qc = res
            res = None
        qs.abs_qc = new_abs_qc
        return res

    def qjit(self, *args, function_name="jaspr_function"):
        """
        Leverages the Catalyst pipeline to compile a QIR representation of
        this function and executes that function using the Catalyst QIR runtime.

        Parameters
        ----------
        *args : iterable
            The arguments to call the function with.

        Returns
        -------
            The values returned by the compiled, executed function.

        """
        flattened_jaspr = self

        from qrisp.jasp.evaluation_tools.catalyst_interface import (
            jaspr_to_catalyst_qjit,
        )

        qjit_obj = jaspr_to_catalyst_qjit(flattened_jaspr, function_name=function_name)
        res = qjit_obj.compiled_function(*args)
        if not isinstance(res, (tuple, list)):
            return res
        elif len(res) == 1:
            return res[0]
        else:
            return res

    @classmethod
    @lru_cache(maxsize=int(1e5))
    def from_cache(cls, closed_jaxpr):
        return Jaspr(jaxpr=closed_jaxpr.jaxpr, consts=closed_jaxpr.consts)

    def update_eqns(self, eqns):
        return Jaspr(
            constvars=list(self.constvars),
            invars=list(self.invars),
            outvars=list(self.outvars),
            eqns=list(eqns),
            consts=list(self.consts)
        )

    def to_qir(self):
        """
        Compiles the Jaspr to QIR using the `Catalyst framework <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The QIR string.

        Examples
        --------

        We create a simple script and inspect the QIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_qir())

        Yields:

        .. code-block:: none

            ; ModuleID = 'LLVMDialectModule'
            source_filename = "LLVMDialectModule"

            @"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}" = internal constant [66 x i8] c"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}\00"
            @lightning.qubit = internal constant [16 x i8] c"lightning.qubit\00"
            @"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so" = internal constant [111 x i8] c"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so\00"

            declare void @__catalyst__rt__finalize() local_unnamed_addr

            declare void @__catalyst__rt__initialize() local_unnamed_addr

            declare ptr @__catalyst__qis__Measure(ptr, i32) local_unnamed_addr

            declare void @__catalyst__qis__T(ptr, ptr) local_unnamed_addr

            declare void @__catalyst__qis__CNOT(ptr, ptr, ptr) local_unnamed_addr

            declare ptr @__catalyst__rt__array_get_element_ptr_1d(ptr, i64) local_unnamed_addr

            declare ptr @__catalyst__rt__qubit_allocate_array(i64) local_unnamed_addr

            declare void @__catalyst__rt__device_init(ptr, ptr, ptr) local_unnamed_addr

            declare void @_mlir_memref_to_llvm_free(ptr) local_unnamed_addr

            declare ptr @_mlir_memref_to_llvm_alloc(i64) local_unnamed_addr

            define { ptr, ptr, i64 } @jit_jaspr_function(ptr nocapture readnone %0, ptr nocapture readonly %1, i64 %2) local_unnamed_addr {
              tail call void @__catalyst__rt__device_init(ptr nonnull @"/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so", ptr nonnull @lightning.qubit, ptr nonnull @"{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}")
              %4 = tail call ptr @__catalyst__rt__qubit_allocate_array(i64 20)
              %5 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 0)
              %6 = load ptr, ptr %5, align 8
              %7 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 1)
              %8 = load ptr, ptr %7, align 8
              tail call void @__catalyst__qis__CNOT(ptr %6, ptr %8, ptr null)
              %9 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 1)
              %10 = load ptr, ptr %9, align 8
              tail call void @__catalyst__qis__T(ptr %10, ptr null)
              %11 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 65)
              %12 = ptrtoint ptr %11 to i64
              %13 = add i64 %12, 63
              %14 = and i64 %13, -64
              %15 = inttoptr i64 %14 to ptr
              %16 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 65)
              %17 = ptrtoint ptr %16 to i64
              %18 = add i64 %17, 63
              %19 = and i64 %18, -64
              %20 = inttoptr i64 %19 to ptr
              %21 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %22 = ptrtoint ptr %21 to i64
              %23 = add i64 %22, 63
              %24 = and i64 %23, -64
              %25 = inttoptr i64 %24 to ptr
              %26 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %27 = ptrtoint ptr %26 to i64
              %28 = add i64 %27, 63
              %29 = and i64 %28, -64
              %30 = inttoptr i64 %29 to ptr
              %31 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %32 = ptrtoint ptr %31 to i64
              %33 = add i64 %32, 63
              %34 = and i64 %33, -64
              %35 = inttoptr i64 %34 to ptr
              %36 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %37 = ptrtoint ptr %36 to i64
              %38 = add i64 %37, 63
              %39 = and i64 %38, -64
              %40 = inttoptr i64 %39 to ptr
              %41 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              store i64 0, ptr %41, align 1
              %42 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              store i64 0, ptr %42, align 1
              %43 = load i64, ptr %1, align 4
              %44 = icmp slt i64 %43, 1
              store i1 %44, ptr %15, align 64
              %45 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %46 = load i64, ptr %42, align 1
              store i64 %46, ptr %45, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %42)
              %47 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %48 = load i64, ptr %41, align 1
              store i64 %48, ptr %47, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %41)
              br i1 %44, label %.lr.ph, label %._crit_edge

            .lr.ph:                                           ; preds = %3, %.lr.ph
              %49 = phi ptr [ %87, %.lr.ph ], [ %47, %3 ]
              %50 = phi ptr [ %85, %.lr.ph ], [ %45, %3 ]
              %51 = load i64, ptr %50, align 4
              %52 = tail call ptr @__catalyst__rt__array_get_element_ptr_1d(ptr %4, i64 %51)
              %53 = load ptr, ptr %52, align 8
              %54 = tail call ptr @__catalyst__qis__Measure(ptr %53, i32 -1)
              %55 = load i1, ptr %54, align 1
              store i1 %55, ptr %20, align 64
              %56 = load i64, ptr %50, align 4
              store i64 %56, ptr %25, align 64
              %57 = shl i64 2, %56
              %58 = icmp ult i64 %56, 64
              %59 = select i1 %58, i64 %57, i64 0
              store i64 %59, ptr %30, align 64
              %60 = load i1, ptr %20, align 64
              %61 = zext i1 %60 to i64
              store i64 %61, ptr %35, align 64
              %62 = load i64, ptr %30, align 64
              %63 = select i1 %60, i64 %62, i64 0
              store i64 %63, ptr %40, align 64
              %64 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %65 = ptrtoint ptr %64 to i64
              %66 = add i64 %65, 63
              %67 = and i64 %66, -64
              %68 = inttoptr i64 %67 to ptr
              %69 = load i64, ptr %49, align 4
              %70 = load i64, ptr %40, align 64
              %71 = add i64 %70, %69
              store i64 %71, ptr %68, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %49)
              %72 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 72)
              %73 = ptrtoint ptr %72 to i64
              %74 = add i64 %73, 63
              %75 = and i64 %74, -64
              %76 = inttoptr i64 %75 to ptr
              %77 = load i64, ptr %50, align 4
              %78 = add i64 %77, 1
              store i64 %78, ptr %76, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %50)
              %79 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %80 = load i64, ptr %68, align 64
              store i64 %80, ptr %79, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr %64)
              %81 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %82 = load i64, ptr %76, align 64
              store i64 %82, ptr %81, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr %72)
              %.pre = load i64, ptr %81, align 4
              %83 = load i64, ptr %1, align 4
              %84 = icmp sge i64 %.pre, %83
              store i1 %84, ptr %15, align 64
              %85 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %86 = load i64, ptr %81, align 1
              store i64 %86, ptr %85, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %81)
              %87 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 8)
              %88 = load i64, ptr %79, align 1
              store i64 %88, ptr %87, align 1
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %79)
              br i1 %84, label %.lr.ph, label %._crit_edge

            ._crit_edge:                                      ; preds = %.lr.ph, %3
              %.lcssa20 = phi ptr [ %45, %3 ], [ %85, %.lr.ph ]
              %.lcssa = phi ptr [ %47, %3 ], [ %87, %.lr.ph ]
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %.lcssa20)
              tail call void @_mlir_memref_to_llvm_free(ptr %36)
              tail call void @_mlir_memref_to_llvm_free(ptr %31)
              tail call void @_mlir_memref_to_llvm_free(ptr %26)
              tail call void @_mlir_memref_to_llvm_free(ptr %21)
              tail call void @_mlir_memref_to_llvm_free(ptr %16)
              tail call void @_mlir_memref_to_llvm_free(ptr %11)
              %89 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 68)
              %90 = ptrtoint ptr %89 to i64
              %91 = add i64 %90, 63
              %92 = and i64 %91, -64
              %93 = inttoptr i64 %92 to ptr
              %94 = load i64, ptr %.lcssa, align 4
              %95 = trunc i64 %94 to i32
              store i32 %95, ptr %93, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr nonnull %.lcssa)
              %96 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 68)
              %97 = ptrtoint ptr %96 to i64
              %98 = add i64 %97, 63
              %99 = and i64 %98, -64
              %100 = inttoptr i64 %99 to ptr
              %101 = load i32, ptr %93, align 64
              %102 = add i32 %101, 1
              store i32 %102, ptr %100, align 64
              tail call void @_mlir_memref_to_llvm_free(ptr %89)
              %103 = icmp eq ptr %96, inttoptr (i64 3735928559 to ptr)
              br i1 %103, label %104, label %107

            104:                                              ; preds = %._crit_edge
              %105 = tail call ptr @_mlir_memref_to_llvm_alloc(i64 4)
              %106 = load i32, ptr %100, align 64
              store i32 %106, ptr %105, align 1
              br label %107

            107:                                              ; preds = %104, %._crit_edge
              %.pn16 = phi ptr [ %105, %104 ], [ %96, %._crit_edge ]
              %.pn14 = phi ptr [ %105, %104 ], [ %100, %._crit_edge ]
              %.pn13 = insertvalue { ptr, ptr, i64 } undef, ptr %.pn16, 0
              %.pn = insertvalue { ptr, ptr, i64 } %.pn13, ptr %.pn14, 1
              %108 = insertvalue { ptr, ptr, i64 } %.pn, i64 0, 2
              ret { ptr, ptr, i64 } %108
            }

            define void @_catalyst_pyface_jit_jaspr_function(ptr nocapture writeonly %0, ptr nocapture readonly %1) local_unnamed_addr {
              %.unpack = load ptr, ptr %1, align 8
              %.elt1.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %.unpack, i64 0, i32 1
              %.unpack2.i = load ptr, ptr %.elt1.i, align 8
              %3 = tail call { ptr, ptr, i64 } @jit_jaspr_function(ptr poison, ptr %.unpack2.i, i64 poison)
              %.elt.i = extractvalue { ptr, ptr, i64 } %3, 0
              store ptr %.elt.i, ptr %0, align 8
              %.repack5.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 1
              %.elt6.i = extractvalue { ptr, ptr, i64 } %3, 1
              store ptr %.elt6.i, ptr %.repack5.i, align 8
              %.repack7.i = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 2
              %.elt8.i = extractvalue { ptr, ptr, i64 } %3, 2
              store i64 %.elt8.i, ptr %.repack7.i, align 8
              ret void
            }

            define void @_catalyst_ciface_jit_jaspr_function(ptr nocapture writeonly %0, ptr nocapture readonly %1) local_unnamed_addr {
              %.elt1 = getelementptr inbounds { ptr, ptr, i64 }, ptr %1, i64 0, i32 1
              %.unpack2 = load ptr, ptr %.elt1, align 8
              %3 = tail call { ptr, ptr, i64 } @jit_jaspr_function(ptr poison, ptr %.unpack2, i64 poison)
              %.elt = extractvalue { ptr, ptr, i64 } %3, 0
              store ptr %.elt, ptr %0, align 8
              %.repack5 = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 1
              %.elt6 = extractvalue { ptr, ptr, i64 } %3, 1
              store ptr %.elt6, ptr %.repack5, align 8
              %.repack7 = getelementptr inbounds { ptr, ptr, i64 }, ptr %0, i64 0, i32 2
              %.elt8 = extractvalue { ptr, ptr, i64 } %3, 2
              store i64 %.elt8, ptr %.repack7, align 8
              ret void
            }

            define void @setup() local_unnamed_addr {
              tail call void @__catalyst__rt__initialize()
              ret void
            }

            define void @teardown() local_unnamed_addr {
              tail call void @__catalyst__rt__finalize()
              ret void
            }

            !llvm.module.flags = !{!0}

            !0 = !{i32 2, !"Debug Info Version", i32 3}

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import jaspr_to_qir

        return jaspr_to_qir(self.flatten_environments())
    
    def to_mlir(self):
        """
        Compiles the Jaspr to an xDSL module using the Jasp Dialect.
        Requires the xDSL package to be installed (``pip install xdsl``).
        
        
        .. note::
        
            An xDSL module can be visualized via:
                
            ::
                
                print(xdsl_module)
                
            and serialized to a string using:
                
            ::
                
                from xdsl.printer import Printer
                Printer().print_op(xdsl_module)
        

        Returns
        -------
        xdsl.dialects.builtin.ModuleOp
            An xDSL module representing the quantum computation.
            
        Examples
        --------

        We create a simple script and inspect the MLIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_mlir())
            
        .. code-block:: none
        
            builtin.module @jasp_module {
              func.func public @main(%arg0 : tensor<i64>, %arg1 : !jasp.QuantumState) -> (tensor<i64>, !jasp.QuantumState) {
                %0, %1 = "jasp.create_qubits"(%arg0, %arg1) : (tensor<i64>, !jasp.QuantumState) -> (!jasp.QubitArray, !jasp.QuantumState)
                %2 = "stablehlo.constant"() <{value = dense<0> : tensor<i64>}> : () -> tensor<i64>
                %3 = "jasp.get_qubit"(%0, %2) : (!jasp.QubitArray, tensor<i64>) -> !jasp.Qubit
                %4 = "stablehlo.constant"() <{value = dense<1> : tensor<i64>}> : () -> tensor<i64>
                %5 = "jasp.get_qubit"(%0, %4) : (!jasp.QubitArray, tensor<i64>) -> !jasp.Qubit
                %6 = "jasp.quantum_gate"(%3, %5, %1) {gate_type = "cx"} : (!jasp.Qubit, !jasp.Qubit, !jasp.QuantumState) -> !jasp.QuantumState
                %7 = "jasp.quantum_gate"(%5, %6) {gate_type = "t"} : (!jasp.Qubit, !jasp.QuantumState) -> !jasp.QuantumState
                %8, %9 = "jasp.measure"(%0, %7) : (!jasp.QubitArray, !jasp.QuantumState) -> (tensor<i64>, !jasp.QuantumState)
                %10 = "stablehlo.add"(%8, %4) : (tensor<i64>, tensor<i64>) -> tensor<i64>
                %11 = "jasp.reset"(%0, %9) : (!jasp.QubitArray, !jasp.QuantumState) -> !jasp.QuantumState
                %12 = "jasp.delete_qubits"(%0, %11) : (!jasp.QubitArray, !jasp.QuantumState) -> !jasp.QuantumState
                func.return %10, %12 : tensor<i64>, !jasp.QuantumState
              }
            }
        
        """
        from qrisp.jasp.mlir import jaspr_to_mlir
        return jaspr_to_mlir(self)

    def to_catalyst_mlir(self):
        """
        Compiles the Jaspr to MLIR using the `Catalyst dialect <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The MLIR string.

        Examples
        --------

        We create a simple script and inspect the MLIR string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)
            print(jaspr.to_catalyst_mlir())

        .. code-block:: none

                        module @jaspr_function {
              func.func public @jit_jaspr_function(%arg0: tensor<i64>) -> tensor<i32> attributes {llvm.emit_c_interface} {
                %0 = stablehlo.constant dense<1> : tensor<i32>
                %1 = stablehlo.constant dense<2> : tensor<i64>
                %2 = stablehlo.constant dense<1> : tensor<i64>
                %3 = stablehlo.constant dense<0> : tensor<i64>
                quantum.device["/home/positr0nium/miniconda3/envs/qrisp/lib/python3.10/site-packages/catalyst/utils/../lib/librtd_lightning.so", "lightning.qubit", "{'shots': 0, 'mcmc': False, 'num_burnin': 0, 'kernel_name': None}"]
                %4 = quantum.alloc( 20) : !quantum.reg
                %5 = quantum.extract %4[ 0] : !quantum.reg -> !quantum.bit
                %6 = quantum.extract %4[ 1] : !quantum.reg -> !quantum.bit
                %out_qubits:2 = quantum.custom "CNOT"() %5, %6 : !quantum.bit, !quantum.bit
                %7 = quantum.insert %4[ 0], %out_qubits#0 : !quantum.reg, !quantum.bit
                %8 = quantum.insert %7[ 1], %out_qubits#1 : !quantum.reg, !quantum.bit
                %9 = quantum.extract %8[ 1] : !quantum.reg -> !quantum.bit
                %out_qubits_0 = quantum.custom "T"() %9 : !quantum.bit
                %10 = quantum.insert %8[ 1], %out_qubits_0 : !quantum.reg, !quantum.bit
                %11 = stablehlo.add %3, %arg0 : tensor<i64>
                %12:3 = scf.while (%arg1 = %3, %arg2 = %3, %arg3 = %10) : (tensor<i64>, tensor<i64>, !quantum.reg) -> (tensor<i64>, tensor<i64>, !quantum.reg) {
                  %16 = stablehlo.compare  GE, %arg1, %11,  SIGNED : (tensor<i64>, tensor<i64>) -> tensor<i1>
                  %extracted = tensor.extract %16[] : tensor<i1>
                  scf.condition(%extracted) %arg1, %arg2, %arg3 : tensor<i64>, tensor<i64>, !quantum.reg
                } do {
                ^bb0(%arg1: tensor<i64>, %arg2: tensor<i64>, %arg3: !quantum.reg):
                  %extracted = tensor.extract %arg1[] : tensor<i64>
                  %16 = quantum.extract %arg3[%extracted] : !quantum.reg -> !quantum.bit
                  %mres, %out_qubit = quantum.measure %16 : i1, !quantum.bit
                  %from_elements = tensor.from_elements %mres : tensor<i1>
                  %extracted_1 = tensor.extract %arg1[] : tensor<i64>
                  %17 = quantum.insert %arg3[%extracted_1], %out_qubit : !quantum.reg, !quantum.bit
                  %18 = stablehlo.subtract %arg1, %3 : tensor<i64>
                  %19 = stablehlo.shift_left %1, %18 : tensor<i64>
                  %20 = stablehlo.convert %from_elements : (tensor<i1>) -> tensor<i64>
                  %21 = stablehlo.multiply %19, %20 : tensor<i64>
                  %22 = stablehlo.add %arg2, %21 : tensor<i64>
                  %23 = stablehlo.add %arg1, %2 : tensor<i64>
                  scf.yield %23, %22, %17 : tensor<i64>, tensor<i64>, !quantum.reg
                }
                %13 = stablehlo.convert %12#1 : (tensor<i64>) -> tensor<i32>
                %14 = stablehlo.multiply %13, %0 : tensor<i32>
                %15 = stablehlo.add %14, %0 : tensor<i32>
                return %15 : tensor<i32>
              }
              func.func @setup() {
                quantum.init
                return
              }
              func.func @teardown() {
                quantum.finalize
                return
              }
            }

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import jaspr_to_mlir

        return jaspr_to_mlir(self.flatten_environments())

    def to_qasm(self, *args):
        """
        Compiles the Jaspr into an OpenQASM 2 string. Real-time control is possible
        as long as no computations on the measurement results are performed.

        Parameters
        ----------
        *args : list
            The arguments to call the :ref:`QuantumCircuit` evaluation with.

        Returns
        -------
        str
            The OpenQASM 2 string.

        Examples
        --------

        We create a simple script and inspect the QASM 2 string:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def main(i):

                qv = QuantumVariable(i)
                cx(qv[0], qv[1])
                t(qv[1])
                return qv

            jaspr = make_jaspr(main)(2)

            qasm_str = jaspr.to_qasm(2)
            print(qasm_str)
            # Yields

            # OPENQASM 2.0;
            # include "qelib1.inc";
            # qreg qb_59[1];
            # qreg qb_60[1];
            # cx qb_59[0],qb_60[0];
            # t qb_60[0];


        It is also possible to compile simple real-time control features:

        ::

            def main(phi):

                qf = QuantumFloat(5)
                h(qf)
                bl = measure(qf[0])

                with control(bl):
                    rz(phi, qf[1])
                    x(qf[1])

                return

            jaspr = make_jaspr(main)(0.5)
            print(jaspr.to_qasm(0.5))

        This gives:

        ::

            OPENQASM 2.0;
            include "qelib1.inc";
            qreg qb_59[1];
            qreg qb_60[1];
            qreg qb_61[1];
            qreg qb_62[1];
            qreg qb_63[1];
            creg cb_0[1];
            h qb_59[0];
            h qb_60[0];
            h qb_61[0];
            reset qb_61[0];
            h qb_62[0];
            reset qb_62[0];
            h qb_63[0];
            reset qb_63[0];
            measure qb_59[0] -> cb_0[0];
            reset qb_59[0];
            if(cb_0==1) rz(0.5) qb_60[0];
            if(cb_0==1) x qb_60[0];
            reset qb_60[0];

        """
        res = self.to_qc(*args)
        if len(self.outvars) == 1:
            res = [res]
        qrisp_qc = res[-1]
        return qrisp_qc.qasm()

    def to_catalyst_jaxpr(self):
        """
        Compiles the jaspr to the corresponding `Catalyst jaxpr <https://docs.pennylane.ai/projects/catalyst/en/stable/index.html>`__.

        Parameters
        ----------
        *args : tuple
            The arguments to call the jaspr with.

        Returns
        -------
        jax.core.Jaxpr
            The Jaxpr using Catalyst primitives.

        Examples
        --------

        We create a simple script and inspect the Catalyst Jaxpr:

        ::

            from qrisp import *
            from qrisp.jasp import make_jaspr

            def example_function(i):

                qv = QuantumFloat(i)
                cx(qv[0], qv[1])
                t(qv[1])
                meas_res = measure(qv)
                meas_res += 1
                return meas_res

            jaspr = make_jaspr(example_function)(2)

            print(jaspr.to_catalyst_jaxpr())
            # Yields
            # { lambda ; a:AbstractQreg() b:i64[] c:i32[]. let
            #   d:i64[] = convert_element_type[new_dtype=int64 weak_type=True] c
            #   e:i64[] = add b d
            #   f:i64[] = add b 0
            #   g:i64[] = add b 1
            #   h:AbstractQbit() = qextract a f
            #   i:AbstractQbit() = qextract a g
            #   j:AbstractQbit() k:AbstractQbit() = qinst[op=CNOT qubits_len=2] h i
            #   l:AbstractQreg() = qinsert a f j
            #   m:AbstractQreg() = qinsert l g k
            #   n:AbstractQbit() = qextract m g
            #   o:AbstractQbit() = qinst[op=T qubits_len=1] n
            #   p:AbstractQreg() = qinsert m g o
            #   q:i64[] = convert_element_type[new_dtype=int64 weak_type=True] c
            #   r:i64[] = add b q
            #   _:i64[] s:i64[] t:AbstractQreg() _:i64[] _:i64[] = while_loop[
            #     body_jaxpr={ lambda ; u:i64[] v:i64[] w:AbstractQreg() x:i64[] y:i64[]. let
            #         z:AbstractQbit() = qextract w u
            #         ba:bool[] bb:AbstractQbit() = qmeasure z
            #         bc:AbstractQreg() = qinsert w u bb
            #         bd:i64[] = sub u x
            #         be:i64[] = shift_left 2 bd
            #         bf:i64[] = convert_element_type[new_dtype=int64 weak_type=True] ba
            #         bg:i64[] = mul be bf
            #         bh:i64[] = add v bg
            #         bi:i64[] = add u 1
            #       in (bi, bh, bc, x, y) }
            #     body_nconsts=0
            #     cond_jaxpr={ lambda ; bj:i64[] bk:i64[] bl:AbstractQreg() bm:i64[] bn:i64[]. let
            #         bo:bool[] = ge bj bn
            #       in (bo,) }
            #     cond_nconsts=0
            #     nimplicit=0
            #     preserve_dimensions=True
            #   ] b 0 p b r
            #   bp:i32[] = convert_element_type[new_dtype=int64 weak_type=False] s
            #   bq:i32[] = mul bp 1
            #   br:i32[] = add bq 1
            # in (t, e, br) }

        """
        from qrisp.jasp.evaluation_tools.catalyst_interface import (
            jaspr_to_catalyst_jaxpr,
        )

        return jaspr_to_catalyst_jaxpr(self.flatten_environments())


def make_jaspr(fun, garbage_collection="auto", flatten_envs=True, **jax_kwargs):
    from qrisp.jasp import (
        AbstractQuantumCircuit,
        TracingQuantumSession,
        check_for_tracing_mode,
    )
    from qrisp.core import recursive_qv_search

    def jaspr_creator(*args, **kwargs):

        qs = TracingQuantumSession.get_instance()

        # Close any tracing quantum sessions that might have not been
        # properly closed due to whatever reason.
        if not check_for_tracing_mode():
            while qs.abs_qc is not None:
                qs.conclude_tracing()

        # This function will be traced by Jax.
        # Note that we add the abs_qc keyword as the tracing quantum circuit
        def ammended_function(*args, **kwargs):

            abs_qc = kwargs[10*"~"]
            del kwargs[10*"~"]

            qs.start_tracing(abs_qc, garbage_collection)

            # If the signature contains QuantumVariables, these QuantumVariables went
            # through a flattening/unflattening procedure. The unflattening creates
            # a copy of the QuantumVariable object, which is however not yet registered in any
            # QuantumSession. We register these QuantumVariables in the current QuantumSession.
            arg_qvs = recursive_qv_search(args)
            for qv in arg_qvs:
                qs.register_qv(qv, None)

            try:
                res = fun(*args, **kwargs)
            except Exception as e:
                qs.conclude_tracing()
                raise e

            res_qvs = recursive_qv_search(res)

            qs.garbage_collection(spare_qv_list=arg_qvs + res_qvs)

            res_qc = qs.conclude_tracing()

            return res, res_qc

        ammended_kwargs = dict(kwargs)
        ammended_kwargs[10*"~"] = AbstractQuantumCircuit()
        closed_jaxpr = make_jaxpr(ammended_function, **jax_kwargs)(
            *args, **ammended_kwargs
        )
            
        # Collect the environments
        # This means that the quantum environments no longer appear as
        # enter/exit primitives but as primitive that "call" a certain Jaspr.
        res = Jaspr.from_cache(collect_environments(closed_jaxpr))

        if flatten_envs:
            res = res.flatten_environments()

        return res

    # Since we are calling the "ammended function", where the first parameter
    # is the AbstractQuantumCircuit, we need to move the static_argnums indicator.
    if "static_argnums" in jax_kwargs:
        jax_kwargs = dict(jax_kwargs)
        if isinstance(jax_kwargs["static_argnums"], list):
            jax_kwargs["static_argnums"] = list(jax_kwargs["static_argnums"])
            for i in range(len(jax_kwargs["static_argnums"])):
                jax_kwargs["static_argnums"][i] += 1
        else:
            jax_kwargs["static_argnums"] += 1

    return jaspr_creator


def check_aval_equivalence(invars_1, invars_2):
    avals_1 = [invar.aval for invar in invars_1]
    avals_2 = [invar.aval for invar in invars_2]
    return all([type(avals_1[i]) == type(avals_2[i]) for i in range(len(avals_1))])

class ProcessedMeasurement:
    pass


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.core import QuantumArray, QuantumVariable, x, cx, mcx
from qrisp.qtypes import QuantumBool
from qrisp.environments import conjugate, control, custom_inversion, invert, custom_control
from qrisp.alg_primitives import demux
from qrisp.core.gate_application_functions import mcx
from qrisp.jasp import check_for_tracing_mode, jrange, q_fori_loop, q_cond
import numpy as np
import jax.numpy as jnp


def qswitch(operand, case, case_function, method="auto", case_amount=None, inv=False, ctrl=None):
    r"""
    Executes a switch - case statement distinguishing between a list of
    given in-place functions.

    More precisely, the qswitch applies the unitary $U_i$ to the operand in state $\ket{\psi}$ given that the case variable is in state $\ket{i}$, i.e.,

    .. math:: 

        \text{qswitch}\ket{i}_{\text{case}}\ket{\psi}_{\text{operand}} = \ket{i}_{\text{case}}U_i\ket{\psi}_{\text{operand}}

    Parameters
    ----------
    operand : :ref:`QuantumVariable`
        The argument on which the case function operates.
    case : :ref:`QuantumFloat`
        The index specifying which case should be executed.
    case_function : list[callable] or callable
        A list of functions, performing some in-place operation on ``operand``, or
        a function ``case_function(i, operand)`` performing some in-place operation on ``operand`` depending on a nonnegative integer index ``i`` specifying the case.
    method : str, optional
        The compilation method. Available are ``sequential``, ``parallel``, ``tree`` and ``auto``.
        ``parallel`` is exponentially fast but requires more temporary qubits. ``tree`` uses `balanced binaray trees <https://arxiv.org/pdf/2407.17966v1>`_.
        The default is ``auto``.
    case_amount : int, optional
        The number of cases. By default the number is inferred automatically:
        - When ``case_function`` is a single function, the size of the case variable is used.
        - When ``case_function`` is a list of functions, the length of that list is used instead.

    Examples
    --------

    First, we consider the case where ``case_function`` is a **list of functions**:

    We create some sample functions:

    ::

        from qrisp import *

        def f0(x): x += 1
        def f1(x): inpl_mult(x, 3, treat_overflow = False)
        def f2(x): pass
        def f3(x): h(x[1])
        case_function_list = [f0, f1, f2, f3]

    Create operand and case variable:

    ::

        operand = QuantumFloat(4)
        operand[:] = 1
        case = QuantumFloat(2)
        h(case)

    Execute switch - case function:

    >>> qswitch(operand, case, case_function_list)

    Simulate:

    >>> print(multi_measurement([case, operand]))
    {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}


    Second, we consider the case where ``case_function`` is a **function**:

    ::

        def case_function(i, qv):
            x(qv[i])

        operand = QuantumFloat(4)
        case = QuantumFloat(2)
        h(case)

        qswitch(operand, case, case_function)

    Simulate:

    >>> print(multi_measurement([case, operand]))
    {(0, 1): 0.25, (1, 2): 0.25, (2, 4): 0.25, (3, 8): 0.25}

    """

    def invert_inpl_function(func):
        def return_func(*args, **kwargs):
            with invert():
                res = func(*args, **kwargs)
            return res
        return return_func

    if callable(case_function):
        if case_amount == None:
            case_amount = 2**case.size
        if check_for_tracing_mode():
            xrange = jrange
        else:
            xrange = range
        if method == "auto":
            method = "tree"

        if inv:
            case_function = invert_inpl_function(case_function)

    elif isinstance(case_function, list):
        if case_amount == None:
            case_amount = len(case_function)
        else:
            if method == "sequential":
                raise TypeError(
                    "Argument 'case_amount' must be 'None' when using the 'sequential' method and a list as a 'case_function'"
                    )

        if inv:
            case_function = [invert_inpl_function(func) for func in case_function]

        xrange = range
        if method == "auto":
            method = "tree"

    else:
        raise TypeError(
            "Argument 'case_function' must be a list or a callable(i, x)"
        )

    if method == "sequential":

        control_qbl = QuantumBool()

        for i in xrange(case_amount):
            with conjugate(mcx)(case, control_qbl, ctrl_state=i):
                with control(control_qbl):
                    if ctrl is None:
                        if callable(case_function):
                            case_function(i, operand)
                        else:
                            case_function[i](operand)
                    else:
                        if callable(case_function):
                            with control(ctrl):
                                case_function(i, operand)
                        else:
                            with control(ctrl):
                                case_function[i](operand)

        control_qbl.delete()

    elif method == "parallel":

        if check_for_tracing_mode():
            raise Exception(
                f"Compile method {method} for switch-case structure not available in tracing mode."
            )

        # Idea: Use demux function to move operand and enabling bool into QuantumArray
        # to execute cases in parallel.

        # This QuantumArray acts as an addressable QRAM via the demux function

        if case_amount != 2**case.size:
            import warnings
            warnings.warn(
                "Warning: Additional qubit overhead because case amount is smaller than case QuantumVariable!")

        enable = QuantumArray(qtype=QuantumBool(), shape=(2**case.size,))
        enable[0].flip()

        qa = QuantumArray(qtype=operand, shape=((2**case.size,)))

        with conjugate(demux)(operand, case, qa, parallelize_qc=True):
            with conjugate(demux)(enable[0], case, enable, parallelize_qc=True):
                for i in range(case_amount):
                    with control(enable[i]):
                        if ctrl is None:
                            if callable(case_function):
                                case_function(i, qa[i])
                            else:
                                case_function[i](qa[i])
                        else:
                            if callable(case_function):
                                with control(ctrl):
                                    case_function(i, qa[i])
                            else:
                                with control(ctrl):
                                    case_function[i](qa[i])

        qa.delete()

        enable[0].flip()
        enable.delete()

    # Uses balanced binaray trees https://arxiv.org/pdf/2407.17966v1
    elif method == "tree":

        # Jasp mode
        if check_for_tracing_mode():
            xrange = jrange
            x_fori_loop = q_fori_loop
            x_cond = q_cond

            def bitwise_count_diff(a, b):
                return jnp.int32(jnp.bitwise_count(jnp.bitwise_xor(a, b)))

        # Normal mode
        else:
            xrange = range

            def x_fori_loop(lower, upper, body_fun, init_val):
                val = init_val
                for i in range(lower, upper):
                    val = body_fun(i, val)
                return val

            def x_cond(pred, true_fun, false_fun, *operands):
                if pred:
                    return true_fun(*operands)
                else:
                    return false_fun(*operands)

            def bitwise_count_diff(a, b):
                return np.int32(np.bitwise_count(np.bitwise_xor(a, b)))

        n = case.size

        def nor_x(t): x(t)
        def nor_cx(c, t): cx(c, t)
        def nor_mcx(c, t): mcx(c, t)

        def bounce(d: int, anc, ca, oper):
            # with control(anc[d - 2]):
            #    x(anc[d-1])
            if ctrl is None:
                x_cond(d - 2 == -1, lambda: nor_x(anc[d-1]),
                       lambda: nor_cx(anc[d-2], anc[d-1]))
            else:
                x_cond(d - 2 == -1, lambda: nor_cx(ctrl,
                       anc[d-1]), lambda: nor_cx(anc[d-2], anc[d-1]))

            with control(anc[d-1]):
                x(anc[d])

            # with control(anc[d - 2]):
            #    with control(ca[n - 1 - d]):
            #        x(anc[d])
            if ctrl is None:
                x_cond(d - 2 == -1, lambda: nor_cx(ca[n - 1 - d], anc[d]),
                       lambda: nor_mcx([ca[n - 1 - d], anc[d-2]], anc[d]))
            else:
                x_cond(d - 2 == -1, lambda: nor_mcx([ca[n - 1 - d], ctrl],
                       anc[d]), lambda: nor_mcx([ca[n - 1 - d], anc[d-2]], anc[d]))

        def down(d: int, anc, ca, oper):
            x(ca[n - 1 - d])
            # with control(anc[d-1]):
            #    with control(ca[n - 1 - d]):
            #        x(anc[d])
            if ctrl is None:
                x_cond(d - 1 == -1, lambda: nor_cx(ca[n - 1 - d], anc[d]),
                       lambda: nor_mcx([ca[n - 1 - d], anc[d-1]], anc[d]))
            else:
                x_cond(d - 1 == -1, lambda: nor_mcx([ca[n - 1 - d], ctrl],
                       anc[d]), lambda: nor_mcx([ca[n - 1 - d], anc[d-1]], anc[d]))
            x(ca[n - 1 - d])

        def up(d: int, anc, ca, oper):
            # with control(anc[d-1]):
            #    with control(ca[n - 1 - d]):
            #        x(anc[d])
            if ctrl is None:
                x_cond(d - 1 == -1, lambda: nor_cx(ca[n - 1 - d], anc[d]),
                       lambda: nor_mcx([ca[n - 1 - d], anc[d-1]], anc[d]))
            else:
                x_cond(d - 1 == -1, lambda: nor_mcx([ca[n - 1 - d], ctrl],
                       anc[d]), lambda: nor_mcx([ca[n - 1 - d], anc[d-1]], anc[d]))

        # Function mode
        if callable(case_function):

            def leaf(d: int, anc, ca, oper, i):
                with control(anc[d]):
                    case_function(i, oper)

                # with control(anc[d-1]):
                #    x(anc[d])
                if ctrl is None:
                    x_cond(d-1 == -1, lambda: nor_x(anc[d]),
                           lambda: nor_cx(anc[d-1], anc[d]))
                else:
                    x_cond(d-1 == -1, lambda: nor_cx(ctrl,
                           anc[d]), lambda: nor_cx(anc[d-1], anc[d]))

                with control(anc[d]):
                    case_function(i + 1, oper)

            def last_leaf(d: int, anc, ca, oper, i):
                with control(anc[d]):
                    case_function(i, oper)

        # List mode
        elif isinstance(case_function, list):

            if len(case_function) % 2 != 0:
                def identity(_):
                    pass

                case_function.append(identity)

            if check_for_tracing_mode():

                def leaf(d: int, anc, ca, oper, i):
                    def apply_leaf(A, B):
                        with control(anc[d]):
                            A(oper)

                        # with control(anc[d-1]):
                        #    x(anc[d])
                        if ctrl is None:
                            x_cond(
                                d-1 == -1, lambda: nor_x(anc[d]), lambda: nor_cx(anc[d-1], anc[d]))
                        else:
                            x_cond(d-1 == -1, lambda: nor_cx(ctrl,
                                   anc[d]), lambda: nor_cx(anc[d-1], anc[d]))

                        with control(anc[d]):
                            B(oper)

                    for j in range(0, len(case_function), 2):
                        x_cond(
                            j == i,
                            apply_leaf,
                            lambda a, b: None,
                            case_function[j],
                            case_function[j + 1],
                        )

            else:

                def leaf(d: int, anc, ca, oper, i):
                    with control(anc[d]):
                        case_function[i](oper)

                    # with control(anc[d-1]):
                    #    x(anc[d])
                    if ctrl is None:
                        x_cond(
                            d-1 == -1, lambda: nor_x(anc[d]), lambda: nor_cx(anc[d-1], anc[d]))
                    else:
                        x_cond(d-1 == -1, lambda: nor_cx(ctrl,
                               anc[d]), lambda: nor_cx(anc[d-1], anc[d]))

                    with control(anc[d]):
                        case_function[i + 1](oper)

            def last_leaf(d: int, anc, ca, oper, i):
                def apply(f):
                    with control(anc[d]):
                        f(oper)
                for j in range(0, len(case_function)):
                    x_cond(j == i, apply, lambda x: None, case_function[j])

        else:
            raise TypeError(
                "Argument 'case_function' must be a list or a callable(i, x)"
            )

        def body_fun(pos, val):
            anc, ca, oper = val

            # Apply leaf
            leaf(n - 1, anc, ca, oper, 2 * pos)

            # Jump to next leaf
            q = bitwise_count_diff(pos, pos + 1)
            for j in xrange(0, q - 1, 1):
                up(n - j - 1, anc, ca, oper)
            bounce(n - q, anc, ca, oper)
            for j in xrange(0, q - 1, 1):
                down(n - (q - 1) + j, anc, ca, oper)

            return anc, ca, oper

        anc = QuantumVariable(n)
        # x(anc[0])

        # Go to first node
        for j in xrange(0, n, 1):
            down(j, anc, case, operand)

        # Perform leafs and jumps

        _, _, _ = x_fori_loop(
            0, -(-case_amount // 2) - 1, body_fun, (anc, case, operand)
        )

        # Perfrom last leaf
        x_cond(case_amount % 2 == 0,
               lambda: leaf(n - 1, anc, case, operand, case_amount - 2),
               lambda: last_leaf(n - 1, anc, case, operand, case_amount - 1))

        # Go back from last node
        diff = 2**n - case_amount
        for j in xrange(0, n, 1):
            up(n - j - 1, anc, case, operand)

            def bf():
                # with control(anc[n-j-2]):
                #    x(anc[n - j-1])
                # return None
                if ctrl is None:
                    x_cond(
                        n-j-2 == -1, lambda: nor_x(anc[n-j-1]), lambda: nor_cx(anc[n-j-2], anc[n-j-1]))
                else:
                    x_cond(n-j-2 == -1, lambda: nor_cx(ctrl,
                           anc[n-j-1]), lambda: nor_cx(anc[n-j-2], anc[n-j-1]))

            # The x_cond applies:
            # if (diff >> j) & 1:
            #    with control(anc[n-j-1]):
            #        x(anc[n - j])
            x_cond((diff >> j) & 1, lambda: bf(), lambda: None)

        # x(anc[0])

        anc.delete()

    else:
        raise Exception(
            f"Don't know compile method {method} for switch-case structure."
        )


temp = qswitch.__doc__
qswitch = custom_control(custom_inversion(qswitch))
qswitch.__doc__ = temp



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import jax.numpy as jnp

from qrisp import check_for_tracing_mode
from qrisp.qtypes.quantum_float import QuantumFloat
from qrisp.misc import gate_wrap
from qrisp.core import QuantumVariable
import jax


def comparison_wrapper(func):

    def res_func(self, other):

        if self.m != 0:
            raise Exception(
                "Tried to evaluate QuantumModulus comparison with non-zero Montgomery shift"
            )

        conversion_flag = False
        if isinstance(other, QuantumModulus):

            if other.m != 0:
                raise Exception(
                    "Tried to evaluate QuantumModulus comparison with non-zero Montgomery shift"
                )

            if self.modulus != other.modulus:
                raise Exception(
                    "Tried to compare QuantumModulus instances of differing modulus"
                )

            other.__class__ = QuantumFloat
        self.__class__ = QuantumFloat
        res = func(self, other)
        self.__class__ = QuantumModulus
        if conversion_flag:
            other.__class__ = QuantumModulus
        return res

    return res_func


class QuantumModulus(QuantumFloat):
    r"""
    This class is a subtype of :ref:`QuantumFloat`, which can be used to model and
    process `modular arithmetic <https://en.wikipedia.org/wiki/Modular_arithmetic>`_.
    Modular arithmetic plays an important role in many cryptographical applications,
    especially in Shor's algorithm.

    The QuantumModulus allows users convenient access to modular arithmetic, which
    for many gate based frameworks is rather complicated due to the intricacies
    of the reduction process.

    For a first example we simply add two instances:

    >>> from qrisp import QuantumModulus
    >>> N = 13
    >>> a = QuantumModulus(N)
    >>> b = QuantumModulus(N)
    >>> a[:] = 5
    >>> b[:] = 10
    >>> c = a + b
    >>> print(c)
    {2: 1.0}

    We get the output 2 because

    >>> (5 + 10)%13
    2

    Similar to :ref:`QuantumFloat`, subtraction and addition are also supported:

    >>> d = a*b
    >>> print(d)
    {11: 1.0}

    Check the result:

    >>> (5*10)%13
    11

    Especially relevant for Shor's algorithm are the in-place operations:

    >>> a = QuantumModulus(N)
    >>> a[:] = 5
    >>> a *= 10
    >>> print(a)
    {11: 1.0}

    **Specifying a custom adder**

    It is possible to specify a custom adder that is used when processing the
    Modular-Arithmetic. For this, you can use the ``inpl_adder`` keyword.

    By default, the `Fourier-adder <https://arxiv.org/abs/quant-ph/0008033>`_
    is used, but we can for instance also try the `Cuccaro-adder <https://arxiv.org/abs/quant-ph/0410184>`_.

    >>> from qrisp import cuccaro_adder
    >>> a = QuantumModulus(N, inpl_adder = cuccaro_adder)
    >>> a[:] = 5
    >>> a *= 10
    >>> print(a)
    {11: 1.0}

    Or the `Gidney-adder <https://arxiv.org/abs/1709.06648>`_.

    >>> from qrisp import gidney_adder
    >>> a = QuantumModulus(N, inpl_adder = gidney_adder)
    >>> a[:] = 5
    >>> a *= 10
    >>> print(a)
    {11: 1.0}

    To learn how to create your own adder for this feature, please visit :meth:`this page <qrisp.inpl_adder_test>`.

    .. warning::

        Currently the adder is only used in in-place multiplication, since this is the
        relevant operation for Shor's algorithm. The other operations (such as addition etc.)
        will follow in a future release of Qrisp.

    **Advanced usage**

    The modular multiplication uses a technique called `Montgomery reduction <https://en.wikipedia.org/wiki/Montgomery_modular_multiplication>`_.
    The quantum version of this algorithm can be found in `this paper <https://arxiv.org/abs/1801.01081>`__.
    The idea behind Montgomery reduction is to choose a differing representation of numbers to enhance
    the reduction step of modular arithmetic. This representation works as follows:
    For an integer $m$ called Montgomery shift, the modular number $a \in \mathbb{Z}/N\mathbb{Z}$ is represented as

    .. math::

        \hat{k} = (2^{-m} k) \text{mod} N

    If you're interested in why this representation is advantageous, we recommend
    checking out the linked resources above.

    For Qrisp, the Montgomery shift can be modified via the attribute ``m``.

    >>> a = QuantumModulus(N)
    >>> a.m = 3
    >>> a[:] = 1
    >>> print(a)
    {1: 1.0}

    We shift back to 0:

    >>> a.m -= 3
    >>> print(a)
    {8: 1.0}

    Note that this shift is only a compiler shift - ie. no quantum gates are applied.
    Instead the :meth:`decoder <qrisp.QuantumVariable.decoder>` function is modified.

    """

    def __init__(self, modulus, inpl_adder=None, qs=None):

        if check_for_tracing_mode():
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import smallest_power_of_two
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import BigInteger
            #if isinstance(modulus, BigInteger):
            #    pad = jnp.zeros(modulus.digits.shape[0], dtype=modulus.digits.dtype)
            #    self.modulus_double = BigInteger(jnp.concatenate([modulus.digits, pad], axis=0))
            self.modulus = modulus
            aux = smallest_power_of_two(modulus)

            QuantumFloat.__init__(self, msize=aux, qs=qs)
            if inpl_adder is None:
                from qrisp.alg_primitives.arithmetic import gidney_adder
                inpl_adder = gidney_adder

            self.inpl_adder = inpl_adder

            self.m = 0

        else:
            self.modulus = modulus
            self.m = int(np.ceil(np.log2(modulus)))

            QuantumFloat.__init__(self, msize=self.m, qs=qs)

            if inpl_adder is None:
                from qrisp.alg_primitives.arithmetic import fourier_adder

                inpl_adder = fourier_adder

            self.inpl_adder = inpl_adder

            self.m = 0

    def decoder(self, i):
        if check_for_tracing_mode():
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import (
                BigInteger
            )
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import (
                montgomery_decoder
            )

            if isinstance(i, BigInteger):
                n = i.digits.shape[0]
                R = BigInteger.create(1, n) << self.m
                return montgomery_decoder(i, R, self.modulus)
            
            else:
                from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import montgomery_decoder
                return montgomery_decoder(i, 2**self.m, self.modulus)
        

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import (
            montgomery_decoder,
        )

        if i >= self.modulus:  # or (np.gcd(i, self.modulus) != 1 and i != 0):
            return np.nan
        return montgomery_decoder(i, 2 ** self.m, self.modulus)
    
    def jdecoder(self, i):
        from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import montgomery_decoder
        return montgomery_decoder(i, 2 ** self.m, self.modulus)
    
    def measure(self):
        from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import (
            BigInteger
        )
        if isinstance(self.modulus, BigInteger):
            from qrisp import q_fori_loop, measure, jlen
            if check_for_tracing_mode():
                for_loop = q_fori_loop
            else:
                def for_loop(lower, upper, body_fun, init_val):
                    val = init_val
                    for i in range(lower, upper):
                        val = body_fun(i, val)
                    return val
            def body_fun(i, val):
                return val.at[i].set(measure(self[32*i:32*(i+1)]).astype(jnp.uint32))
            digits = for_loop(0, (self.size-1)//32, body_fun, jnp.zeros_like(self.modulus.digits))
            digits = digits.at[(self.size-1)//32].set(measure(self[32*((self.size - 1)//32):]).astype(jnp.uint32))
            return self.jdecoder(BigInteger(digits))
        else:
            return self.jdecoder(self.reg.measure())

    def encoder(self, i):
        if check_for_tracing_mode():
        
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import (
                BigInteger
            )
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import montgomery_encoder
            if isinstance(i, BigInteger):
                return montgomery_encoder(i, BigInteger.create(1, i.digits.shape[0]) << self.m, self.modulus)
            else:
                return montgomery_encoder(i, 1 << self.m, self.modulus)
        
        else:

            if i >= self.modulus:
                raise Exception(
                    "Tried to encode a number into QuantumModulus, which is greator or equal to the modulus"
                )
            if i < 0:
                raise Exception("Tried to encode a negative number into QuantumModulus")

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import (
            montgomery_encoder,
        )

        # if i >= self.modulus:  # or (np.gcd(i, self.modulus) != 1 and i != 0):
        #     return np.nan

        return montgomery_encoder(i, 2**self.m, self.modulus)

    #def encode(self, i):
    #    QuantumVariable.encode(self, self.encoder(i))

    @gate_wrap(permeability="args", is_qfree=True)
    def __mul__(self, other):
        from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import BigInteger
        if isinstance(other, QuantumModulus):
            if self.m != other.m:
                raise ValueError("Both QuantumModuli must have the same shift")
            if self.modulus != other.modulus:
                raise ValueError("Both QuantumModuli must have the same modulus")
            if check_for_tracing_mode():
                from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_montgomery import (
                    qq_montgomery_multiply_modulus
                )
                return qq_montgomery_multiply_modulus(self, other)
            else:
                from qrisp.alg_primitives.arithmetic.modular_arithmetic import (
                    montgomery_mod_mul,
                )           
                return montgomery_mod_mul(self, other)

        elif isinstance(other, (int, np.integer, jnp.integer, BigInteger)):
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_montgomery import (
                cq_montgomery_multiply
            )
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import (
                best_montgomery_shift
            )
            shift = best_montgomery_shift(other, self.modulus)
            if isinstance(self.modulus, BigInteger):
                assert isinstance(other, BigInteger)
                return cq_montgomery_multiply(other.get_larger(), self, self.modulus.get_larger(), shift)
            return cq_montgomery_multiply(other, self, self.modulus, shift)
        else:
            raise Exception(
                f"Quantum modular multiplication with type {type(other)} not implemented"
            )

    __rmul__ = __mul__

    @gate_wrap(permeability=[1], is_qfree=True)
    def __imul__(self, other):
        from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_bigintiger import BigInteger

        if isinstance(other, (int, np.integer, jnp.integer, jax.Array, BigInteger)):
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_montgomery import (
                cq_montgomery_multiply_inplace
            )
            from qrisp.alg_primitives.arithmetic.jasp_arithmetic.jasp_mod_tools import (
                best_montgomery_shift
            )
            
            shift = best_montgomery_shift(other, self.modulus)
            if isinstance(self.modulus, BigInteger):
                assert isinstance(other, BigInteger)
                cq_montgomery_multiply_inplace(other.get_larger(), self, self.modulus.get_larger(), shift, self.inpl_adder)
            else:
                cq_montgomery_multiply_inplace(other, self, self.modulus, shift, self.inpl_adder)
        
            if isinstance(other, BigInteger):
                pad = jnp.zeros(other.digits.shape[0], dtype=other.digits.dtype)
                other = BigInteger(jnp.concatenate([other.digits, pad], axis=0))
            
            return self
        else: 
            raise Exception(
                f"Quantum modular in-place multiplication with type {type(other)} not implemented"
            )


    @gate_wrap(permeability="args", is_qfree=True)
    def __add__(self, other):
        if isinstance(other, int):
            other = self.encoder(other % self.modulus)
        elif isinstance(other, QuantumModulus):
            if self.m != other.m:
                raise Exception(
                    "Tried to add two QuantumModulus with differing Montgomery shift"
                )
        elif isinstance(other, QuantumFloat):
            if self.m != 0:
                raise Exception(
                    "Tried to add a QuantumFloat and QuantumModulus with non-zero Montgomery shift"
                )

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import mod_adder

        res = self.duplicate(init=True)

        mod_adder(other, res, self.inpl_adder, self.modulus)

        return res

    __radd__ = __mul__

    @gate_wrap(permeability=[1], is_qfree=True)
    def __iadd__(self, other):
        if isinstance(other, int):
            other = self.encoder(other % self.modulus)
        elif isinstance(other, QuantumModulus):
            if self.m != other.m:
                from qrisp.alg_primitives.arithmetic.modular_arithmetic import (
                    montgomery_addition,
                )

                montgomery_addition(other, self)
                return self
        elif isinstance(other, QuantumFloat):
            if self.m != 0:
                raise Exception(
                    "Tried to add a QuantumFloat and QuantumModulus with non-zero Montgomery shift"
                )

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import mod_adder

        mod_adder(other, self, self.inpl_adder, self.modulus)
        return self

    @gate_wrap(permeability="args", is_qfree=True)
    def __sub__(self, other):
        if isinstance(other, int):
            other = self.encoder(other % self.modulus)
        elif isinstance(other, QuantumModulus):
            if self.m != other.m:
                raise Exception(
                    "Tried to add subtract QuantumModulus with differing Montgomery shift"
                )
        elif isinstance(other, QuantumFloat):
            if self.m != 0:
                raise Exception(
                    "Tried to subtract a QuantumFloat and QuantumModulus with non-zero Montgomery shift"
                )

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import mod_adder
        from qrisp.environments import invert

        res = self.duplicate(init=True)

        with invert():
            mod_adder(other, res, self.inpl_adder, self.modulus)

        return res

    @gate_wrap(permeability="args", is_qfree=True)
    def __rsub__(self, other):
        if isinstance(other, int):
            other = self.encoder(other % self.modulus)
        elif isinstance(other, QuantumModulus):
            if self.m != other.m:
                raise Exception(
                    "Tried to add subtract QuantumModulus with differing Montgomery shift"
                )
        elif isinstance(other, QuantumFloat):
            if self.m != 0:
                raise Exception(
                    "Tried to subtract a QuantumFloat and QuantumModulus with non-zero Montgomery shift"
                )

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import mod_adder

        res = self.duplicate()

        res -= self

        mod_adder(other, res, self.inpl_adder, self.modulus)

        return res

    @gate_wrap(permeability=[1], is_qfree=True)
    def __isub__(self, other):
        if isinstance(other, int):
            other = self.encoder(other % self.modulus)
        elif isinstance(other, QuantumModulus):
            if self.m != other.m:
                raise Exception(
                    "Tried to add subtract QuantumModulus with differing Montgomery shift"
                )
        elif isinstance(other, QuantumFloat):
            if self.m != 0:
                raise Exception(
                    "Tried to subtract a QuantumFloat and QuantumModulus with non-zero Montgomery shift"
                )

        from qrisp.alg_primitives.arithmetic.modular_arithmetic import mod_adder
        from qrisp.environments import invert

        with invert():
            mod_adder(other, self, self.inpl_adder, self.modulus)

        return self

    @comparison_wrapper
    def __lt__(self, other):
        from qrisp.alg_primitives import uint_lt

        return uint_lt(self, other, self.inpl_adder)

    @comparison_wrapper
    def __gt__(self, other):
        from qrisp.alg_primitives import uint_gt

        return uint_gt(self, other, self.inpl_adder)

    @comparison_wrapper
    def __le__(self, other):
        from qrisp.alg_primitives import uint_le

        return uint_le(self, other, self.inpl_adder)

    @comparison_wrapper
    def __ge__(self, other):
        from qrisp.alg_primitives import uint_ge

        return uint_ge(self, other, self.inpl_adder)

    @comparison_wrapper
    def __eq__(self, other):
        return QuantumVariable.__eq__(self, other)

    @comparison_wrapper
    def __ne__(self, other):
        return QuantumVariable.__ne__(self, other)

    def __hash__(self):
        return QuantumFloat.__hash__(self)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import sympy as sp
from sympy.polys.polytools import degree

from qrisp.alg_primitives.arithmetic.poly_tools import (
    expr_to_list,
    filter_pow,
    get_ordered_symbol_list,
)
from qrisp.core import (
    QuantumArray,
    QuantumVariable,
    cp,
    cx,
    cz,
    h,
    mcx,
    p,
    z,
    rz,
    rzz,
    crz,
    mcp,
    gphase,
)
from qrisp.misc import gate_wrap, lifted
from qrisp.circuit import XGate

# Threshold of rounding used in detecting integer multiples of pi
pi_mult_round_threshold = 11

depth_tracker = {}


# Check if the given expression is a polynomial
def check_for_polynomial(expr):
    for x in sp.preorder_traversal(expr):
        if not isinstance(
            x,
            (
                sp.core.mul.Mul,
                sp.core.add.Add,
                sp.core.symbol.Symbol,
                sp.core.numbers.Integer,
                sp.core.numbers.Float,
                sp.core.power.Pow,
            ),
        ):
            return False

        if isinstance(x, sp.core.power.Pow):
            if not isinstance(x.args[1], sp.core.numbers.Integer):
                return False
    return True


# Efficient implementation of the multicontrolled U_g gate
def multi_controlled_U_g(
    output_qf, control_qb_list, y, phase_tolerant=False, use_gms=False
):
    # Set alias for quantum session
    # qs = output_qf.qs
    qs = control_qb_list[0].qs()

    # For one control qubit, there is no advantage in calculating an ancilla qubit
    if len(control_qb_list) == 1:
        ancilla = control_qb_list
    # Otherwise request an ancilla qubit
    else:
        ancilla = QuantumVariable(1, qs=output_qf.qs, name="sbp_anc*")

    # If an ancilla qb is available
    # we synthesize the result of the boolean multiplication in it

    if len(control_qb_list) != 1:
        if use_gms:
            toffoli_method = "gms"
        else:
            toffoli_method = "gray_pt"

        # Apply multi-controlled x gate
        mcx(control_qb_list, ancilla[0], method=toffoli_method)

    # Now we execute the phase gates
    # For this there is 2 possibilies
    # Either use regular phase gates or use GMS gates
    # GMS gates are ion-trap native gates, that allow the entanglement
    # of multiple qubits within a single laser pulse

    from qrisp.environments import (
        QuantumEnvironment,
        control,
        GMSEnvironment,
        custom_control,
    )

    if use_gms:
        # GMSEnvironment is an environment that allows programming
        # with regular phase gates but converts everything programmed
        # into a GMS gate upon leaving the environment

        env = GMSEnvironment(qs)

    else:
        # This is an environment that simply does nothing
        env = QuantumEnvironment()

    # Enter environment
    env.manual_allocation_management = True

    @custom_control
    def crz_helper(rot_angle, a, b, ctrl=None, use_gms=False):
        if ctrl is None and not use_gms:
            cx(a, b)
            p(-rot_angle / 2, b)
            cx(a, b)
            p(rot_angle / 2, b)
        elif use_gms:
            crz(rot_angle, ancilla[0], output_qf[i])
        else:
            cx(a, b)
            cp(-rot_angle / 2, ctrl, b)
            cx(a, b)
            cp(rot_angle / 2, ctrl, b)

    with env:
        phase_accumulator = 0
        cz_counter = 0

        # Execute controlled rotations
        for i in range(len(output_qf)):
            # The -i (instead of i) reverses the gate order implying that
            # we can leave out the swaps of the QFT

            rot_angle = 2 * np.pi * 2 ** (-i - 1) * y
            rot_angle = rot_angle % (2 * np.pi)

            if (
                np.round(rot_angle, pi_mult_round_threshold) != 0
                and np.round((-rot_angle) % (2 * np.pi), pi_mult_round_threshold) != 0
            ):
                # If phase is equal to pi, execute cz gate (costs one CNOT less than CP)
                if False:
                    # if np.round(abs(rot_angle) - np.pi, pi_mult_round_threshold) == 0:
                    # cz(ancilla[0], output_qf.reg[i])

                    z(output_qf.reg[i])
                    cz_counter += 1

                    phase_accumulator += rot_angle / 2
                # Otherwise execute cp gate
                else:
                    crz_helper(rot_angle, ancilla[0], output_qf[i], use_gms=use_gms)
                    phase_accumulator += rot_angle / 2

        p(phase_accumulator - np.pi * cz_counter / 2, ancilla[0])

    # Uncompute boolean multiplication
    if len(control_qb_list) != 1:
        if not use_gms:
            toffoli_method += "_inv"

        mcx(control_qb_list, ancilla[0], method=toffoli_method)
        ancilla.delete()

    return phase_accumulator


# This function returns the U_g gate (up to qiskit endian conventions)
# This is used for applying U_g gates which have no control-knobs
def U_g(y, qv):
    for i in range(len(qv)):
        rot_angle = np.pi * 2 ** (-i) * y
        rot_angle = rot_angle % (2 * np.pi)
        if (
            np.round(rot_angle, pi_mult_round_threshold) != 0
            and np.round((-rot_angle) % (2 * np.pi), pi_mult_round_threshold) != 0
        ):
            p(rot_angle, qv[i])


# This function encodes a semi-boolean polynomial into a circuit i.e.
# For the polynomial p(x_0, x_1, x_2) = 4*x_0*x_1 + 2*x_0*x_2
# the effect of this function is:
# U_f |x_0, x_1, x_2>|0> = |x_0, x_1, x_2>|p(x_1,x_2,x_3)>
def sb_polynomial_encoder(
    input_qf_list, output_qf, poly, inplace_mult=1, use_gms=False, init_op="auto"
):
    # As the polynomial has only boolean variables,
    # powers can be ignored since x**k = x for x in GF(2)
    poly = filter_pow(poly.expand()).expand() / 2.0**output_qf.exponent

    # Acquire list of symbols present in the polynomial
    symbol_list = []

    for qf in input_qf_list:
        temp_var_list = list(sp.symbols(str(hash(qf)) + "_" + "0:" + str(qf.size)))
        symbol_list += temp_var_list

    n = len(symbol_list)

    if n != sum([var.size for var in input_qf_list]):
        raise Exception(
            "Input variables do not the required amount of qubits to encode polynomial"
        )

    # Acquire monomials in list form
    monomial_list = expr_to_list(poly)

    from qrisp import QFT, conjugate, QuantumEnvironment

    if inplace_mult == 1:
        env = conjugate(QFT)(
            output_qf,
            inv=False,
            exec_swap=False,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
        )
    else:
        QFT(
            output_qf,
            inv=False,
            exec_swap=False,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
        )
        env = QuantumEnvironment()

    with env:
        # The list of qubits contained in the variables of input_var_list
        input_qubits = sum([list(var.reg) for var in input_qf_list], [])

        control_qubit_list = []
        y_list = []

        # Iterate through the monomials
        for monom in monomial_list:
            # Prepare the two variables coeff (which is the coefficient of the monomial)
            # And the list of variables which appear in the monomial

            # For this, go through the cases which can appear

            # This describes the case where there is only a single term in the monomial
            # Either a constant or a variable
            if len(monom) == 1:
                if isinstance(monom[0], sp.core.symbol.Symbol):
                    coeff = 1
                    variables = list(monom)
                else:
                    coeff = float(monom[0])
                    variables = []

            # This describes the case where there is multiple terms in the monomial
            elif not isinstance(monom[0], sp.core.symbol.Symbol):
                coeff = monom[0]
                variables = list(monom[1:])
            else:
                coeff = 1
                variables = list(monom)

            # Check if the coefficient is an integer (up to float errors)
            if abs(int(np.round(float(coeff))) - coeff) > 1e-14:
                pass
                # raise Exception("Tried to encode sb-polynomial
                # with non-integer coefficient")

            # Append coefficient to y_list
            y_list.append(int(np.round(float(coeff))))

            # Prepare the qubits on which the U_g should be controlled
            control_qubit_numbers = [symbol_list.index(var) for var in variables]

            control_qubits = [input_qubits[nr] for nr in control_qubit_numbers]

            control_qubits = list(set(control_qubits))

            control_qubits.sort(key=lambda x: x.identifier)

            control_qubit_list.append(control_qubits)

        # Now we apply the multi controlled U_g gate
        # Here the order in which they are applied makes a huge difference
        # In order to determine, which U_g to apply next, we evaluate a cost function
        # (which we determined through trial and error)
        # and choose the U_g with the lowest cost

        # Note that for this feature to yield an improvement, the quantum session requires
        # multiple free ancilla qubits to work on
        def delay_cost(depth_array):
            return max(depth_array)

        def find_best_monomial(control_qubit_list, depth_dic):
            delay_cost_list = []

            for i in range(len(control_qubit_list)):
                depth_array = np.array([depth_dic[qb] for qb in control_qubit_list[i]])

                delay_cost_list.append(delay_cost(depth_array))

            return np.argmin(delay_cost_list)

        # Iterate through the list of U_g gates
        while control_qubit_list:
            # TO-DO fix depth calculation inside environment
            # Update depth_dic (contains the depth of each qubit)
            # depth_dic = output_qf.qs.get_depth_dic()

            # Determine best U_g
            # monomial_index = find_best_monomial(control_qubit_list, depth_dic)
            monomial_index = 0
            # Find control qubits and their coefficient
            control_qubits = control_qubit_list.pop(monomial_index)
            y = y_list.pop(monomial_index)

            # Apply (controlled) U_g
            if len(control_qubits):
                multi_controlled_U_g(output_qf, control_qubits, y, use_gms=use_gms)
            else:
                U_g(y, output_qf)

    if inplace_mult != 1:
        # Apply QFT
        QFT(output_qf, inv=True, exec_swap=False, use_gms=use_gms)


# Multiplies two integers
# The general idea is to represent the integers as values of two polynomials
# p(x) = x_0 + 2*x_1 + 4*x_2 ...
# p(y) = y_0 + 2*y_1 + 4*y_2 ...
# these polynomials get multiplied and which forms a bigger polynomial
# which can be encoded using the polynomial encoder
def sbp_mult(factor_1_qf, factor_2_qf, output_qf=None):
    """
    Performs multiplication based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    factor_1_qf : QuantumFloat
        The first factor to multiply.
    factor_2_qf : QuantumFloat
        The second factor to multiply.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the multiplication.

    Examples
    --------

    We multiply two QuantumFloats:

    ::

        from qrisp import QuantumFloat, sbp_mult
        qf_0 = QuantumFloat(3)
        qf_1 = QuantumFloat(3)
        qf_0[:] = 3
        qf_1[:] = 4
        qf_res = sbp_mult(qf_0, qf_1)
        print(qf_res)


    ::

        #Yields: {12: 1.0}


    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([factor_1_qf, factor_2_qf], op="mul")
    # Multiply the polynmials
    mult_poly = factor_1_qf.sb_poly(output_qf.msize) * factor_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([factor_1_qf, factor_2_qf], output_qf, mult_poly)

    return output_qf


def sbp_add(summand_1_qf, summand_2_qf, output_qf=None):
    """
    Performs addition based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    summand_1_qf : QuantumFloat
        The first summand to add.
    summand_2_qf : QuantumFloat
        The second summand to add.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the addition.

    Examples
    --------

    We add two QuantumFloats:

    ::

        from qrisp import QuantumFloat, sbp_add
        qf_0 = QuantumFloat(3)
        qf_1 = QuantumFloat(3)
        qf_0[:] = 3
        qf_1[:] = 4
        qf_res = sbp_add(qf_0, qf_1)
        print(qf_res)


     ::

        # Yields: {7: 1.0}


    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([summand_1_qf, summand_2_qf], op="add")

    sum_poly = summand_1_qf.sb_poly(output_qf.msize) + summand_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([summand_1_qf, summand_2_qf], output_qf, sum_poly)

    return output_qf


def sbp_sub(summand_1_qf, summand_2_qf, output_qf=None):
    """
    Performs subtraction based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    summand_1_qf : QuantumFloat
        The QuantumFloat to subtract from.
    summand_2_qf : QuantumFloat
        The QuantumFloat to subtract.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the subtraction.

    Examples
    --------

    We add two QuantumFloats:

    ::

        from qrisp import QuantumFloat, sbp_sub
        qf_0 = QuantumFloat(3)
        qf_1 = QuantumFloat(3)
        qf_0[:] = 3
        qf_1[:] = 4
        qf_res = sbp_sub(qf_0, qf_1)
        print(qf_res)


    ::

        # Yields: {-1: 1.0}


    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([summand_1_qf, summand_2_qf], op="sub")

    dif_poly = summand_1_qf.sb_poly(output_qf.msize) - summand_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([summand_1_qf, summand_2_qf], output_qf, dif_poly)

    return output_qf


# Encode the polynomial given in poly in output var
# depending on the input variables qv_list
@gate_wrap(is_qfree=True, permeability=[0])
def polynomial_encoder(qf_list, output_qf, poly, encoding_dic=None, inplace_mult=1):
    """
    Evaluates a (multivariate) sympy polynomial on a list of QuantumFloats using
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    qf_list : list[QuantumFloat]
        The list of QuantumFloats to evaluate the polynomial on.
    output_qf : QuantumFloat
        The QuantumFloat to evaluate into.
    poly : sympy expression
        The polynomial to evaluate.
    encoding_dic : dict, optional
        A dictionary which has the QuantumFloats of qf_list as keys and the associated
        sympy symbols as values. By default, the symbols of the polynomial
        will be ordered alphabetically and then matched to the order in qf_list.
    inplace_mult : int, optional
        This integer allow to perform an inplace multiplication on output_qf,
        before the polynomial is evaluated.
        Note that due to reversibility only odd numbers are supported.

    Raises
    ------
    Exception
        Provided QuantumFloat list does not include the appropriate amount of elements
        to encode given polynomial".

    Returns
    -------
    None.

    Examples
    --------

    We evaluate the polynomial $x^2 + 2y^2$ on two QuantumFloats:


    ::

        from sympy import Symbol
        x = Symbol("x")
        y = Symbol("y")
        poly = x**2 + 2*y**2
        from qrisp import QuantumFloat, polynomial_encoder
        x_qf = QuantumFloat(3)
        y_qf = QuantumFloat(3)
        x_qf[:] = 3
        y_qf[:] = 2
        res_qf = QuantumFloat(7)
        encoding_dic = {x_qf : x, y_qf : y}
        polynomial_encoder([x_qf, y_qf], res_qf, poly, encoding_dic)
        print(res_qf)


    ::

        # Yields: {17.0: 1.0}


    """

    if isinstance(qf_list, QuantumArray):
        qf_list = list(qf_list.flatten())

    if encoding_dic is not None:
        symbol_list = [encoding_dic[qv] for qv in qf_list]

    else:
        symbol_list = get_ordered_symbol_list(poly)

    if len(symbol_list) != len(qf_list):
        raise Exception(
            "Provided QuantumFloat list does not include the appropriate amount"
            "of elements to encode given polynomial"
        )

    if not output_qf.signed:
        for qf in qf_list:
            if qf.signed:
                raise Exception(
                    "When encoding into an unsigned quantum float"
                    "provide only unsigned inputs"
                )

    sb_poly_list = [qf.sb_poly(output_qf.size) for qf in qf_list]

    repl_dic = {symbol_list[i]: sb_poly_list[i] for i in range(len(qf_list))}
    # Substitute SB-polynomials
    sb_polynomial = poly.subs(repl_dic).expand()

    # Apply sb encoder
    sb_polynomial_encoder(qf_list, output_qf, sb_polynomial, inplace_mult=inplace_mult)


def split(a, n):
    k, m = divmod(len(a), n)
    return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))


# Performs inplace addition on a QFT-ed variable
# by applying successive U_g gates
def U_g_inpl_adder(modified_var, summand, mult_factor=1):
    applied_phases = []
    # Extract the coefficients of the SB-polynomial of x
    from sympy import Poly

    summand_coeffs = list(Poly(summand.sb_poly(modified_var.msize)).as_dict().values())
    summand_coeffs = [float(coeff) for coeff in summand_coeffs][::-1]
    # summand_coeffs.sort()

    summand_coeffs = summand_coeffs + (summand.size - len(summand_coeffs)) * [0]

    for j in range(summand.size):
        if summand_coeffs[j] == 0:
            continue
        applied_phases.append(
            multi_controlled_U_g(
                modified_var,
                [summand[j]],
                mult_factor * summand_coeffs[j] * 2**-modified_var.exponent,
            )
        )

    return applied_phases


# This algorithm is based on the equation X*U_g(y)*X = U_g(-y)*G with G as global phase
# This is used to perform conditional addition/subtraction required for multiplication
# using algorithm 3 from https://arxiv.org/abs/2112.10537
# s = x << (n + 1)
# s− = x
# for i in range(y.size):
#   if y[i]:
#       s+= (x << i)
#   else:
#       s−= (x << i)
# return s >> 1
# We will use the function U_g_inply_adder for the inplace additions
# while performing conditional additions using CNOT gates.
# The approach is therefore a hybrid of SBP-method and traditional logic approaches
def hybrid_mult(
    x,
    y,
    output_qf=None,
    init_op="h",
    terminal_op="qft",
    phase_tolerant=False,
    cl_factor=1,
):
    """
    An advanced algorithm for multiplication which has better depth, gate-count
    and compile time than :meth:`sbp_mult <qrisp.sbp_mult>`.
    It does not support squaring a single QuantumFloat though.

    This algorithm also operates on the Fourier transform. Because of this,
    between successive multiplications targeting the same QuantumFloat
    it is not neccessary to Fourier-Transform.
    This advantage is expressed in the parameters init_op and terminal_op.
    These can be set to either 'h', 'qft' or None
    to leave out self canceling Fourier-transforms.

    Parameters
    ----------
    x : QuantumFloat
        The first factor to multiply.
    y : QuantumFloat
        The second factor to multiply.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default a suited QuantumFloat is created.
    init_op : str, optional
        The operation to bring output_qf into it's Fourier-transform.
        The default is 'h'.
    terminal_op : str, optional
        The operation to bring output_qf back from it's Fourier-transform.
        The default is "qft".
    phase_tolerant : bool, optional
        If set to True, differing results introduce differing extra phases.
        This can be usefull to save resources incase this functions will get uncomputed.
        The default is False.
    cl_factor : float, optional
        Allows to multiply the result by a classical factor without any extra gates.
        The default is 1.

    Returns
    -------
    output_qf : QuantumFloat
        The QuantumFloat containing the result.

    Examples
    --------

    We multiply two QuantumFloat with eachother and an additional classical factor

    ::

        from qrisp import QuantumFloat, hybrid_mult
        qf_0 = QuantumFloat(3)
        qf_1 = QuantumFloat(3)
        qf_0[:] = 3
        qf_1[:] = 4
        qf_res = hybrid_mult(qf_0, qf_1, cl_factor = 2)
        print(qf_res)


    ::

        # Yields: {24: 1.0}


    """

    from qrisp import QFT, cx, h, merge, z

    # The two factors take asymetrical roles in this algorithm
    # This implies that there is most likely a prefered choice
    # for the roles depending on the size of the factors
    # Several trials showed that these roles work best
    if not x.size > y.size:
        x, y = y, x

    # We shift the exponent of both factors such that they are integers
    # and shift the result in the end, by the sum of both exponents
    # This allows convenient treatment of non-integer inputs
    # while only constructing the algorithm for integers
    x_exp = int(x.exponent)
    y_exp = int(y.exponent)

    x.exp_shift(-x_exp)
    y.exp_shift(-y_exp)

    # If no output_qf is given, create one
    from qrisp.qtypes.quantum_float import create_output_qf

    if isinstance(output_qf, type(None)):
        output_qf = create_output_qf([y, x], op="mul")

    else:
        output_qf.exp_shift(-(x_exp + y_exp))

    output_qf.extend(1, position=0)

    # output_qf.exp_shift(-1)
    # Since the result is right shifted at the end, this implies that the zero-th qubit
    # of the output will not contain any information (otherwise this would imply,
    # that integer multiplications can yield non-integer results).
    # However the result still needs to be able to
    # display every possible result. This is why it needs "extra working-space"
    # We therefore extend it by one qubit

    # Merge the sessions of all involved variables
    merge([output_qf, x, y])

    # Perform initial operation
    # (check the general documentation of SBP arithmetic for more details)
    if init_op == "h":
        h(output_qf)
    elif init_op == "qft":
        QFT(output_qf, exec_swap=False)
    else:
        h(output_qf[0])

    # We treat the case that y is signed by applying a negation on output_qf that is
    # conditioned on the sign qubit of y (see command [2]). This will be reversed
    # at the end of the algorithm. This implies that every addition that happens
    # in between is actually a subtraction.
    # So this acts as a sign reversal of the result.
    # We then need to make sure we multiply with the absolute of y
    # We do this by applying a bitwise negation on the mantissa of y conditioned on
    # the sign qubit (command [4]). This bitwise negation acts as
    # NOT y = -y + 1
    # This implies we need to correct the additional +1.
    # This additional +1 is equivalent to an extra +x in the result
    # Therefore we need to subtract this +x from output_qf.
    # Note that this subtraction also has to be performed
    # conditioned on the sign qubit of y.
    # Instead of adding an additional control qubit, we again follow the strategy
    # of turning a conditional subtraction into a conditional subtraction/addition
    if y.signed:
        # This performs the first part of the conditional subtraction/addition
        # for j in range(x.size):
        # multi_controlled_U_g(output_qf, [x[j]], -x_coeffs[j])
        U_g_inpl_adder(output_qf, x, -1 * cl_factor)  # command [1]

        # Negate output_qf conditioned on the sign qubit of y
        cx(y[-1], output_qf)  # command [2]

        # This would perform the second part of the subtraction/addition
        # However this command is merged into command [5] for more performance gains.
        # If commented in, this command reverses command [1]
        # if output_qf has not been negated IF the sign qubit
        # has not negated output_qf. Otherwise the subtraction of x is completed.
        # for j in range(x.size):
        # multi_controlled_U_g(output_qf, [x[j]], x_coeffs[j]) #command [3]

        # Negate mantissa of y
        cx(y[-1], y[:-1])  # command [4]

    # This performs the initial two lines of the initially mentioned
    # multiplication algorithm:
    # s = x << (n + 1)
    # s− = x

    # Note that the boolean y.signed adds the phase that would have been added
    # in command [3]
    # however not requiring another round of U_g gates
    applied_phases = U_g_inpl_adder(
        output_qf, x, cl_factor * (2 ** (y.msize) - 1 + y.signed)
    )

    # We now come to the loop of the multiplication algorithm

    # for i in range(y.size):
    #   if y[i]:
    #       s+= (x << i)
    #   else:
    #       s−= (x << i)

    # Negate output_qf in order to perform the conditional addition/subtraction

    if not phase_tolerant and terminal_op != "qft":
        hybrid_mult_anc = QuantumVariable(1)
        if y.signed:
            cx(y[-1], hybrid_mult_anc[0])

            for k in range(len(applied_phases)):
                # cp(-applied_phases[k] * 2, hybrid_mult_anc[0], x[k])
                cp(-applied_phases[k] * 2, x[k], hybrid_mult_anc[0])

    cx(y[0], output_qf)
    for i in range(y.msize):
        # Perform in-place addition
        # (note that if y[i] is active an addition has to be performed).
        # Therefore, we need to perform a subtraction here,
        # because if y[i] is active output_qf has been negated
        applied_phases = U_g_inpl_adder(output_qf, x, -(2**i) * cl_factor)

        if not phase_tolerant and terminal_op != "qft":
            if i != 0:
                cx(y[i - 1], hybrid_mult_anc[0])
            cx(y[i], hybrid_mult_anc[0])

            for k in range(len(applied_phases)):
                # cp(-applied_phases[k] * 2, hybrid_mult_anc[0], x[k])
                cp(-applied_phases[k] * 2, x[k], hybrid_mult_anc[0])
                # sbp+= -2*applied_phases[k]*Symbol("y_" + str(i))*Symbol("x_" + str(k))

        # This command is equivalent to
        # cx(y[i], output_qf)
        # cx(y[i+1], output_qf)
        # ie. performs the output_qf negation but requires less cnot gates
        if i != y.msize - 1:
            cx(y[i], y[i + 1])
            cx(y[i + 1], output_qf)
            cx(y[i], y[i + 1])
        else:
            if not y.signed:
                # This command performs the final negation
                # if y is signed, we can perform the negation
                # together with the negation conditioned on the sign qubit
                # of y (command [6])
                cx(y[i], output_qf)  # command [5]

    if not phase_tolerant and terminal_op != "qft":
        cx(y[i], hybrid_mult_anc[0])
        if y.signed:
            cx(y[-1], hybrid_mult_anc[0])

        # from sympy import Symbol, simplify
        # temp_x = sum([2**i*Symbol("x_" + str(i)) for i in range(3)], 0)
        # temp_y = sum([2**i*Symbol("y_" + str(i)) for i in range(3)], 0)

        # sbp = sbp/(2*np.pi)

        # sbp = simplify((sbp + (temp_x*temp_y).expand()/2**7))
        # print(sbp)
        hybrid_mult_anc.delete()

    if y.signed:
        # This makes sure the negation from command [5] is performed
        cx(y[i], y[-1])

        cx(y[-1], output_qf)  # command[6]

        cx(y[i], y[-1])

        # Reverse command [4]
        cx(y[-1], y[:-1])

    # Free up the qubit which we identified as containing no information

    h(output_qf[0])
    output_qf.reduce(output_qf[0])

    # Perform terminal qft
    if terminal_op == "qft":
        QFT(output_qf, inv=True, exec_swap=False)

    if not phase_tolerant and terminal_op == "qft":
        # This is a phase tolerant correction using only single qubit gates.
        # We found it by looking at the sbp of the correction using cp gates and
        # identified the sbp of x*y in it.
        for i in range(output_qf.size):
            p(-(2 * np.pi) * 2**i / 2 ** (output_qf.size + 1), output_qf[i])

        if output_qf.signed:
            z(output_qf[-1])

    output_qf.exp_shift((x_exp + y_exp))

    # Perform exponent shifts

    x.exp_shift(x_exp)
    y.exp_shift(y_exp)

    return output_qf


# Wrapper for choosing the best multiplication algorithm
def q_mult(factor_1, factor_2, target=None, method="auto"):
    if method == "auto":
        if factor_1.reg == factor_2.reg:
            return q_mult(factor_1, factor_2, target, method="sbp")
        else:
            return q_mult(factor_1, factor_2, target, method="hybrid")

    elif method == "sbp":
        from qrisp.qtypes.quantum_float import create_output_qf

        if target is None:
            target = create_output_qf([factor_1, factor_2], op="mul")

        sbp_mult(factor_1, factor_2, target)
        return target

    elif method == "hybrid":
        from qrisp.alg_primitives.arithmetic import hybrid_mult

        return hybrid_mult(factor_1, factor_2)


def QFT_inpl_mult(qv, inplace_mult=1):
    from qrisp.misc import is_inv

    qv = list(qv)
    qv = qv[::-1]
    n = len(qv)

    if not is_inv(inplace_mult, n):
        raise Exception(
            "Tried to perform non-invertible inplace multiplication during Fourier-Transform"
        )

    # Perform QFT with inplace multiplication
    for i in range(n):
        if i != n - 1:
            h(qv[i])

        if i == n - 1:
            break

        for k in range(n - i - 1):
            if k + i + 1 != n - 1:
                cp(inplace_mult * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])
            else:
                # The -1 here cancels some cp gates of the inverse qft
                cp((inplace_mult - 1) * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])

    # Perform reversed QFT without inplace multiplication and without canceled steps
    for i in range(n)[::-1]:
        for k in range(n - i - 1)[::-1]:
            if k + i + 1 != n - 1:
                cp(-2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])

        if i != n - 1:
            h(qv[i])

    return qv


def inpl_mult(qf, mult_int, treat_overflow=True):
    """
    Performs inplace multiplication of a :ref:`QuantumFloat` with a classical integer.
    To prevent overflow errors, this function automatically adjusts the mantissa size.
    If you want to prevent this behavior, set ``treat_overflow = False``.

    Parameters
    ----------
    qf : QuantumFloat
        The QuantumFloat to inplace-multiply.
    mult_int : int
        The integer to perform the multiplication with.
    treat_overflow : bool
        If set to ``False``, the mantissa will not be extended to prevent overflow errors. The default is ``True``.

    Examples
    --------

    We create a QuantumFloat, bring it to superposition and perform an inplace multiplication.

    ::

        from qrisp import QuantumFloat, h, inpl_mult
        a = QuantumFloat(5, signed = True)
        h(a[0])
        h(a[-1])
        print(a)


    ::

        # Yields: {0: 0.25, 1: 0.25, -32: 0.25, -31: 0.25}


    ::

        inpl_mult(a, -5)
        print(a)


    ::

        # Yields: {0: 0.25, 155: 0.25, 160: 0.25, -5: 0.25}


    """

    if not isinstance(mult_int, (int, float)):
        raise Exception("Quantum inplace multiplication is restricted to classical values due to reversibility constraints")

    if mult_int < 0 and not qf.signed:
        raise Exception(
            "Tried to inplace-multiply unsigned QuantumFloat with negative factor"
        )

    bit_shift = 0

    if int(mult_int) != mult_int:

        c = abs(mult_int)

        for i in range(32):
            if int(2**i * c) == 2**i * c:
                break
        else:
            raise Exception(
                "Tried to inplace multiply with number of to much precision"
            )

        bit_shift = -i
        mult_int = 2**i * mult_int

    else:
        while not mult_int % 2:
            bit_shift += 1
            mult_int = mult_int // 2

    if mult_int != 1:
        if treat_overflow:
            extension_size = int(np.ceil(np.log2(abs(mult_int))))
            qf.extend(extension_size, position=qf.size - 1)
            if qf.signed:
                cx(qf[-1], qf[-1 - extension_size : -1])

        from qrisp.alg_primitives.arithmetic.SBP_arithmetic import QFT_inpl_mult

        QFT_inpl_mult(qf, inplace_mult=mult_int)

    quantum_bit_shift(qf, bit_shift, treat_overflow)

    if treat_overflow and bit_shift < 0 and qf.signed:
        cx(qf[-1], qf[bit_shift - 1 : -1])


def quantum_bit_shift(qf, bit_shift, treat_overflow=True):

    from qrisp import cyclic_shift, control, QuantumFloat

    if isinstance(bit_shift, QuantumFloat):

        if bit_shift.signed or qf.signed:
            raise Exception(
                "Quantum-quantum bitshifting is currently only supported for unsigned arguments"
            )

        for i in range(*bit_shift.mshape):
            with control(bit_shift.significant(i)):
                quantum_bit_shift(qf, 2**i)

        return

    if treat_overflow:

        if bit_shift > 0:

            if qf.signed:
                qf.extend(bit_shift, position=qf.size - 1)
            else:
                qf.extend(bit_shift, position=qf.size)

        else:
            qf.extend(abs(bit_shift), position=0)
            qf.exp_shift(bit_shift)

    if qf.signed:
        cyclic_shift(qf[:-1], bit_shift)
    else:
        cyclic_shift(qf, bit_shift)


# @lifted
def app_sb_phase_polynomial(qv_list, poly, symbol_list=None, t=1):
    """
    Applies a phase function specified by a `semi-Boolean polynomial <https://ieeexplore.ieee.org/document/9815035>`_ acting on a list of QuantumVariables.
    That is, this method implements the transformation

    .. math::

        \ket{y_1}\dotsb\ket{y_n}\\rightarrow e^{itP(y_1,\dotsc,y_n)}\ket{y_1}\dotsb\ket{y_n}

    where :math:`\ket{y_1},\dotsc,\ket{y_n}` are QuantumVariables and :math:`P(y_1,\dotsc,y_n)=P(y_{1,1},\dotsc,y_{1,m_1},\dotsc,y_{n,1}\dotsc,y_{n,m_n})` is a semi-Boolean polynomial in variables
    :math:`y_{1,1},\dotsc,y_{1,m_1},\dotsc,y_{n,1}\dotsc,y_{n,m_n}`. Here, $m_i$ is the size of the $i$ th variable.

    Parameters
    ----------
    qv_list : list[QuantumVariable] or QuantumArray
        The list of QuantumVariables to evaluate the semi-Boolean polynomial on.
    poly : SymPy expression
        The semi-Boolean polynomial to evaluate.
    symbol_list : list, optional
        An ordered list of SymPy symbols associated to the qubits of the QuantumVariables of ``qv_list``.
        For each QuantumVariable in ``qv_list`` a number of symbols according to its size is required.
        By default, the symbols of the polynomial
        will be ordered alphabetically and then matched to the order in ``qv_list``.
    t : Float or SymPy expression, optional
        The argument ``t`` in the expression $\exp(itP)$. The default is 1.

    Raises
    ------
    Exception
        Provided QuantumVariable list does not include the appropriate amount
        of elements to evaluate the given polynomial.

    Examples
    --------

    We apply the phase function specified by the polynomial :math:`P(x,y,z) = \pi xyz` on a QuantumVariable:

    ::

        import sympy as sp
        import numpy as np
        from qrisp import QuantumVariable, app_sb_phase_polynomial

        x, y, z = sp.symbols('x y z')
        P = np.pi*x*y*z

        qv = QuantumVariable(3)
        qv.init_state({'000': 0.5, '111': 0.5})

        app_sb_phase_polynomial([qv], P)

    We print the ``statevector``:

    >>> print(qv.qs.statevector())
    sqrt(2)*(|000> - |111>)/2

    """

    if isinstance(qv_list, QuantumArray):
        qv_list = list(qv_list.flatten())

    # As the polynomial has only boolean variables,
    # powers can be ignored since x**k = x for x in GF(2)
    poly = filter_pow(poly.expand()).expand()

    if symbol_list is None:
        symbol_list = get_ordered_symbol_list(poly)

    if len(symbol_list) != sum([var.size for var in qv_list]):
        raise Exception(
            "Provided QuantumVariable list does not include the appropriate amount "
            "of elements to evaluate the given polynomial"
        )

    # The list of qubits contained in the variables of input_var_list
    input_qubits = sum([list(var.reg) for var in qv_list], [])

    # Monomials in list form
    monomial_list = expr_to_list(poly)

    control_qubit_list = []
    y_list = []

    # Iterate through the monomials
    for monom in monomial_list:
        # Prepare coeff (coefficient of the monomial) and variables (list of variables from symbol_list in the monomial)
        # Note: coeff may also contain symbolic variables
        coeff = float(1)
        variables = []
        for term in monom:
            if isinstance(term, sp.core.symbol.Symbol) and term in symbol_list:
                variables.append(term)
            elif isinstance(term, sp.core.symbol.Symbol):
                coeff = coeff * term
            else:
                coeff = coeff * float(term)

        # Append coefficient to y_list
        y_list.append(coeff)

        # Prepare the qubits on which the phase gate should be controlled
        control_qubit_numbers = [symbol_list.index(var) for var in variables]

        control_qubits = [input_qubits[nr] for nr in control_qubit_numbers]

        control_qubits = list(set(control_qubits))

        control_qubits.sort(key=lambda x: x.identifier)

        control_qubit_list.append(control_qubits)

    # Now we apply the multi controlled phase gates
    # Iterate through the list of phase gates
    while control_qubit_list:
        monomial_index = 0
        # Find control qubits and their coefficient
        control_qubits = control_qubit_list.pop(monomial_index)
        y = y_list.pop(monomial_index)

        # Apply (controlled) phase gate
        if len(control_qubits):
            mcp(y * t, control_qubits)
        else:
            gphase(y * t, input_qubits[0])


# @lifted
def app_phase_polynomial(qf_list, poly, symbol_list=None, t=1):
    """
    Applies a phase function specified by a polynomial acting on a list of QuantumFloats.
    That is, this method implements the transformation

    .. math::

        \ket{y_1}\dotsb\ket{y_n}\\rightarrow e^{itP(y_1,\dotsc,y_n)}\ket{y_1}\dotsb\ket{y_n}

    where :math:`\ket{y_1},\dotsc,\ket{y_n}` are QuantumFloats and :math:`P(y_1,\dotsc,y_n)` is a polynomial in variables
    :math:`y_1,\dotsc,y_n`.

    Parameters
    ----------
    qf_list : list[QuantumFloat] or QuantumArray[QuantumFloat]
        The list of QuantumFloats to evaluate the polynomial on.
    poly : SymPy expression
        The polynomial to evaluate.
    symbol_list : list, optional
        An ordered list of SymPy symbols associated to the QuantumFloats of ``qf_list``.
        By default, the symbols of the polynomial
        will be ordered alphabetically and then matched to the order in ``qf_list``.
    t : Float or SymPy expression, optional
        The argument ``t`` in the expression $\exp(itP)$. The default is 1.

    Raises
    ------
    Exception
        Provided QuantumFloat list does not include the appropriate amount of elements
        to encode given polynomial.

    Examples
    --------

    We apply the phase function specified by the polynomial :math:`P(x,y) = \pi x + \pi xy` on two QuantumFloats:

    ::

        import sympy as sp
        import numpy as np
        from qrisp import QuantumFloat, h, app_phase_polynomial

        x, y = sp.symbols('x y')
        P = np.pi*x + np.pi*x*y

        qf1 = QuantumFloat(3, signed = False)
        qf2 = QuantumFloat(3,-1, signed = False)
        h(qf1[0])
        qf2[:]=0.5

        app_phase_polynomial([qf1,qf2], P)

    We print the ``statevector``:

    >>> print(qf1.qs.statevector())
    sqrt(2)*(|0>*|0.5> - I*|1>*|0.5>)/2


    We apply the phase function specified by the polynomial :math:`P(x) = 1 - 0.9x^2 + x^3` on a QuantumFloat:

    ::

        import numpy as np
        import sympy as sp
        import matplotlib.pyplot as plt
        from qrisp import QuantumFloat, h, app_phase_polynomial

        x = sp.symbols('x')
        P = 1-0.9*x**2+x**3

        qf = QuantumFloat(3,-3)
        h(qf)

        app_phase_polynomial([qf],P)

    To visualize the results we retrieve the ``statevector`` as a function and determine the phase of each entry.

    ::

        sv_function = qf.qs.statevector("function")

    This function receives a dictionary of QuantumVariables specifiying the desired label constellation and returns its complex amplitude.
    We calculate the phases corresponding to the complex amplitudes, and compare the results with the values of the function $P(x)$.

    ::

        qf_values = np.array([qf.decoder(i) for i in range(2 ** qf.size)])
        sv_phase_array = np.angle([sv_function({qf : i}) for i in qf_values])

        P_func = sp.lambdify(x, P, 'numpy')
        x_values = np.linspace(0, 1, 100)
        y_values = P_func(x_values)

    Finally, we plot the results.

    ::

        plt.plot(x_values, y_values, label = "P(x)")
        plt.plot(qf_values , sv_phase_array%(2*np.pi), "o", label = "Simulated phases")
        plt.ylabel("Phase [radian]")
        plt.xlabel("QuantumFloat outcome labels")
        plt.grid()
        plt.legend()
        plt.show()

    .. figure:: /_static/PhasePolynomialApplication.png
        :alt: PhasePolynomialApplication
        :scale: 80%
        :align: center

    """

    if isinstance(qf_list, QuantumArray):
        qf_list = list(qf_list.flatten())

    if symbol_list is None:
        symbol_list = get_ordered_symbol_list(poly)

    if len(symbol_list) != len(qf_list):
        raise Exception(
            "Provided QuantumFloat list does not include the appropriate amount "
            "of elements to evaluate the given polynomial"
        )

    sb_poly_list = []
    new_symbol_list = []
    for qf in qf_list:
        if qf.signed:
            # We do not use modular arithmetic.
            sb_poly_list.append(
                qf.sb_poly()
                - 2 ** (qf.msize + 2 + qf.exponent)
                * sp.symbols(str(hash(qf)) + "_" + str(qf.msize))
            )
        else:
            sb_poly_list.append(qf.sb_poly())

        temp_var_list = list(sp.symbols(str(hash(qf)) + "_" + "0:" + str(qf.size)))
        new_symbol_list += temp_var_list

    repl_dic = {symbol_list[i]: sb_poly_list[i] for i in range(len(qf_list))}
    # Substitute semi-Boolean polynomials
    sb_polynomial = poly.subs(repl_dic).expand()
    # As the polynomial has only boolean variables,
    # powers can be ignored since x**k = x for x in GF(2)
    sb_polynomial = filter_pow(sb_polynomial.expand()).expand()

    # Apply sb phase polynomial
    app_sb_phase_polynomial(qf_list, sb_polynomial, symbol_list=new_symbol_list, t=t)


# Workaround to keep the docstring but still gatewrap

temp = app_sb_phase_polynomial.__doc__

app_sb_phase_polynomial = gate_wrap(permeability="args", is_qfree=True)(
    app_sb_phase_polynomial
)

app_sb_phase_polynomial.__doc__ = temp


temp = app_phase_polynomial.__doc__

app_phase_polynomial = gate_wrap(permeability="args", is_qfree=True)(
    app_phase_polynomial
)

app_phase_polynomial.__doc__ = temp



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.core import Qubit
from qrisp.qtypes import QuantumFloat, QuantumBool
from qrisp.environments import control, invert, conjugate
from qrisp.core.gate_application_functions import x, cx, mcx
from qrisp.jasp import jrange, jnp
from typing import Tuple


def q_max(a: QuantumFloat, b: QuantumFloat) -> QuantumFloat:
    """
    Computes the maximum of two QuantumFloats ``a`` and ``b``.
    
    Parameters
    ----------
    a : QuantumFloat
    b : QuantumFloat

    Returns
    -------
    QuantumFloat
        The maximum value between ``a`` and ``b``.

    Examples
    --------
    >>> from qrisp import *
    >>> a = QuantumFloat(2)
    >>> b = QuantumFloat(2)
    >>> a += 2
    >>> h(a[0])
    >>> b+=1
    >>> res_max = q_max(a,b)
    >>> multi_measurement([a,b,res_max])
    {(2, 1, 2): 0.5, (3, 1, 3): 0.5}
    """

    res_mshape = [
        jnp.minimum(a.mshape[0], b.mshape[0]),
        jnp.maximum(a.mshape[1], b.mshape[1]),
    ]
    new_msize = res_mshape[1] - res_mshape[0]
    new_exponent = res_mshape[0]
    res = QuantumFloat(new_msize, new_exponent, signed=a.signed | b.signed)

    # We calculate the overlap of bits with the same significance of a and b.
    l = jnp.maximum(a.mshape[0], b.mshape[0])
    r = jnp.minimum(a.mshape[1], b.mshape[1])

    c = QuantumBool()
    
    compare_func = lambda x, y: x >= y
    injecteted_comp_func = c << compare_func

    with conjugate(injecteted_comp_func)(a, b):
        cx(a[l - a.exponent : r - a.exponent], res[l - new_exponent : r - new_exponent])

        with conjugate(cx)(
            a[l - a.exponent : r - a.exponent], b[l - b.exponent : r - b.exponent]
        ):
            with control(c, 0):
                cx(
                    b[l - b.exponent : r - b.exponent],
                    res[l - new_exponent : r - new_exponent],
                )

        with control(c):
            up_bd = jnp.minimum(a.size,l - a.exponent)
            cx(a[:up_bd], res[:up_bd])
            lw_bd = jnp.maximum(r, a.exponent)
            cx(a[lw_bd - a.exponent : a.msize], res[lw_bd - new_exponent : a.mshape[1] - new_exponent])
            if a.signed:
                cx(a.sign(), res.sign())

        with control(c, 0):
            up_bd = jnp.minimum(b.size,l - b.exponent)
            cx(b[: up_bd], res[: up_bd])
            lw_bd = jnp.maximum(r, b.exponent)
            cx(b[lw_bd - b.exponent : b.msize], res[lw_bd - new_exponent : b.mshape[1] - new_exponent])
            if b.signed:
                cx(b.sign(), res.sign())

    c.delete()

    return res

def q_min(a: QuantumFloat, b: QuantumFloat) -> QuantumFloat:
    """
    Computes the minimum of two QuantumFloats ``a`` and ``b``.
    
    Parameters
    ----------
    a : QuantumFloat
    b : QuantumFloat

    Returns
    -------
    QuantumFloat
        The minimum value between ``a`` and ``b``.

    Examples
    -------- 
    >>> from qrisp import *
    >>> a = QuantumFloat(2)
    >>> b = QuantumFloat(2)
    >>> a += 2
    >>> h(a[0])
    >>> b+=1
    >>> res_min = q_min(a,b)
    >>> multi_measurement([a,b,res_min])
    {(2, 1, 1): 0.5, (3, 1, 1): 0.5}
    """
    res_mshape = [
        jnp.minimum(a.mshape[0], b.mshape[0]),
        jnp.maximum(a.mshape[1], b.mshape[1]),
    ]
    new_msize = res_mshape[1] - res_mshape[0]
    new_exponent = res_mshape[0]
    res = QuantumFloat(new_msize, new_exponent, signed=a.signed | b.signed)

    # We calculate the overlap of bits with the same significance of a and b.
    l = jnp.maximum(a.mshape[0], b.mshape[0])
    r = jnp.minimum(a.mshape[1], b.mshape[1])

    c = QuantumBool()

    compare_func = lambda x, y: x <= y
    injecteted_comp_func = c << compare_func

    with conjugate(injecteted_comp_func)(a, b):
        cx(a[l - a.exponent : r - a.exponent], res[l - new_exponent : r - new_exponent])

        with conjugate(cx)(
            a[l - a.exponent : r - a.exponent], b[l - b.exponent : r - b.exponent]
        ):
            with control(c, 0):
                cx(
                    b[l - b.exponent : r - b.exponent],
                    res[l - new_exponent : r - new_exponent],
                )

        with control(c):
            up_bd = jnp.minimum(a.size,l - a.exponent)
            cx(a[:up_bd], res[:up_bd])
            lw_bd = jnp.maximum(r, a.exponent)
            cx(a[lw_bd - a.exponent : a.msize], res[lw_bd - new_exponent : a.mshape[1] - new_exponent])
            if a.signed:
                cx(a.sign(), res.sign())

        with control(c, 0):
            up_bd = jnp.minimum(b.size,l - b.exponent)
            cx(b[: up_bd], res[: up_bd])
            lw_bd = jnp.maximum(r, b.exponent)
            cx(b[lw_bd - b.exponent : b.msize], res[lw_bd - new_exponent : b.mshape[1] - new_exponent])
            if b.signed:
                cx(b.sign(), res.sign())

    c.delete()

    return res

def q_floor(a: QuantumFloat) -> QuantumFloat:
    """
    Computes out-of-place the floor of a QuantumFloat.
    
    Parameters
    ----------
    a : QuantumFloat

    Returns
    -------
    QuantumFloat
        The floor of ``a``.

    Examples
    --------
    >>> from qrisp import *
    >>> a = QuantumFloat(4,-2)
    >>> a[:] = {0.25: 0.25**0.5, 1.75: 0.75**0.5}
    >>> b = q_floor(a)
    >>> b.get_measurement()
    {1.0: 0.75, 0.0: 0.25}
    """

    b = a.duplicate()
    for i in jrange(-a.exponent, a.size):
        cx(a[i], b[i])

    return b


def q_ceil(a: QuantumFloat) -> QuantumFloat:
    """
    Computes out-of-place the ceiling of a QuantumFloat.
    
    Parameters
    ----------
    a : QuantumFloat

    Returns
    -------
    QuantumFloat
        The ceiling of ``a``.

    Examples
    --------
    >>> from qrisp import *
    >>> a = QuantumFloat(4,-2)
    >>> a[:] = {0.25: 0.25**0.5, 1.75: 0.75**0.5}
    >>> b = q_ceil(a)
    >>> b.get_measurement()
    {2.0: 0.75, 1.0: 0.25}

    .. warning::
        Ceiling operations that would result in overflow, raise no errors. Instead, the operations
        are performed `modular <https://en.wikipedia.org/wiki/Modular_arithmetic>`_.
    """
    b = q_floor(a)

    with control(a.exponent < 0):
        flag = QuantumFloat(1)

        # Logical OR on the fractional part
        logical_OR(a[:-a.exponent], flag[0])

        with control(flag[0]):
            b += 1

        logical_OR(a[:-a.exponent], flag[0])
        flag.delete()
    return b


def logical_OR(operands: list[Qubit], flag: Qubit):
    # handle the case fo an empty list
    # Implement the case of a Quantum Variable with more the 2 qubits
    x(operands)
    x(flag)
    mcx(operands, flag)
    x(operands)

def q_round(a: QuantumFloat) -> QuantumFloat:
    """
    Computes out-of-place the rounding of a QuantumFloat to the first significant digit.
    
    Parameters
    ----------
    a : QuantumFloat

    Returns
    -------
    QuantumFloat
       The rounding of ``a``.

    Examples
    --------

    >>> from qrisp import *
    >>> a = QuantumFloat(4,-2)
    >>> a[:] = {0.25: 0.25**0.5, 1.75: 0.75**0.5}
    >>> b = q_round(a)
    >>> b.get_measurement()
    {2.0: 0.75, 0.0: 0.25}

    .. warning::
        Rounding operations that would result in overflow, raise no errors. Instead, the operations
        are performed `modular <https://en.wikipedia.org/wiki/Modular_arithmetic>`_.
    """
    b = q_floor(a)

    with control(a.exponent < 0):
        with control(a[-a.exponent - 1]): 
            #Control on the first fractional bit (the one holding the value b*0.5)
            #specifies if the number is rounded up or down.
            b += 1

    return b


def q_fractional(a: QuantumFloat) -> QuantumFloat:
    """
    Computes out-of-place the fractional part of a QuantumFloat.
    
    Parameters
    ----------
    a : QuantumFloat

    Returns
    -------
    QuantumFloat
       The fractional part of ``a``.

    Examples
    --------

    >>> from qrisp import *
    >>> a = QuantumFloat(4,-2)
    >>> a[:] = {0.25: 0.25**0.5, 1.75: 0.75**0.5}
    >>> b = q_fractional(a)
    >>> b.get_measurement()
    {0.75: 0.75, 0.25: 0.25}

    """
    b = a.duplicate()
    for i in jrange(a.size):
        cx(a[i], b[i])

    with control(a.signed):
        flag = QuantumFloat(1)
        cx(a[-1], flag)
        with control(flag): 
            #If the QuantumFloat is signed and the value negative: substract the ceiling
            c = q_ceil(a)
            b -= c
            injected_qceil = c << q_ceil
            with invert():
                injected_qceil(a)
            c.delete()
            
        with control(flag, 0): 
            #If the QuantumFloat is signed and the value positive: substract the floor
            c = q_floor(a)
            b -= c
            injected_qfloor = c << q_floor
            with invert():
                injected_qfloor(a)
            c.delete()
        cx(a[-1], flag)
        flag.delete()
        
    with control(not a.signed):  
        #If the QuantumFloat is unsigned: substract the floor
        c = q_floor(a)
        b -= c
        injected_qfloor = c << q_floor
        with invert():
            injected_qfloor(a)
        c.delete()
    return b


def q_modf(a: QuantumFloat) -> Tuple[QuantumFloat, QuantumFloat]:
    """
    Computes out-of-place the integer and fractional parts of a QuantumFloat.

    Parameters
    ----------
    a : QuantumFloat

    Returns
    -------
    Tuple[QuantumFloat, QuantumFloat]
        A pair `(i, f)` where:

        - `i =` :func:`q_floor(a) <qrisp.q_floor>` is the integer part
        - `f =` :func:`q_fractional(a) <qrisp.q_fractional>` is the fractional part

        Both are new QuantumFloats with the same configuration as ``a``.

    Examples
    --------
    >>> from qrisp import *
    >>> a = QuantumFloat(4, -2)
    >>> a[:] = {0.25: 0.25**0.5, 1.75: 0.75**0.5}
    >>> i, f = qmodf(a)
    >>> i.get_measurement()
    {1.0: 0.75, 0.0: 0.25}
    >>> f.get_measurement()
    {0.75: 0.75, 0.25: 0.25}
    """
    return (q_floor(a), q_fractional(a))
